ai = {}

--needed for ai2_select
function ai2_evaluate(whoami)
	--local bailiff_left = CASTLE_BAILIFF_THRESHOLDS[3] - state.bailiff 
	--local castle_left = CASTLE_CONTRIBUTION_THRESHOLDS[3] - state.contribution_count

	if state.castle_phase <= 3 then
		local me = state.players[whoami]
		local sum = me.score
				+ 4 * me.resources.gold
				+ me.resources.food
				+ me.resources.wood
				+ me.resources.stone
				+ me.resources.cloth
		-- value of money in the bank
		if me.resources.money <= 5 then
			sum = sum + 1 * me.resources.money
		else
			--money not worth as much once you have more than 5
			sum = sum + 1 * 5 + .5 * (me.resources.money - 5)
		end
		-- value of going early in the turn order
		sum = sum + (1.5 - .5 * position_in_turn_order(whoami))
		-- workers on the road
		assert(state.provost > BRIDGE_POSITION)
		local min_provost = math.max(state.provost - 4, BRIDGE_POSITION + 1)
		for i = 1,math.min(state.provost + 3, CASTLE_POSITION - 1) do
			local value_if_activated = 0
			if state.buildings[i].occupants[1] == whoami then
				if i == JOUST_POSITION then
					if me.resources.cloth >= 1 and me.resources.money >= 1 then
						value_if_activated = 3
					end
				elseif i ~= BRIDGE_POSITION then
					--guestimate worth of a building
					value_if_activated = 3
				end
			end
			if i > min_provost then
				local xx = (state.provost + 3 - i) / 6
				assert(xx <= 1.0)
				value_if_activated = value_if_activated * xx
			end
			sum = sum + value_if_activated
		end
		-- workers in the castle
		for i = 1,#state.buildings[CASTLE_POSITION].occupants do
			if state.buildings[CASTLE_POSITION].occupants[i] == whoami then
				-- can player build any blocks?
				local rc = state.players[whoami].resources
				local num_nonfood = rc.wood + rc.stone + rc.cloth
				local can_build_blocks = rc.food >= 1
					and ((rc.wood >= 1 and rc.stone >= 1)
						or (rc.wood >= 1 and rc.cloth >= 1)
						or (rc.stone >= 1 and rc.stone >= 1))
				if can_build_blocks then
					local npoints = 1 + (i == 1 and 1 or 0)
							+ 3 * math.min(num_nonfood, rc.food)
					sum = sum + npoints
				else
					sum = sum - 2
				end
			end
		end

		return sum
	else
		return state.players[whoami].score
	end
end

--needed for several of the AI personalities
function ai2_select()

	local whoami = current_turn()
	local immediate_moves = all_available_moves()
	if #immediate_moves == 1 then
		return immediate_moves[1]
	end

	local immediate_scores = {}
	for i,move in ipairs(immediate_moves) do
		local saved_state = state
		state = clone_game_state(saved_state)
		do_move(move)
		immediate_scores[move] = ai2_evaluate(whoami)
		--print(string.format("move %s gives score of %f", move, immediate_scores[move]))
		state = saved_state
	end

	table.sort(immediate_moves,
		function(a,b)
			return immediate_scores[b] < immediate_scores[a]
		end
		)
	return immediate_moves[1]
end
function play_ai_game_quiet()
	while state.castle_phase <= 3 and state.turn ~= 0 do
		if state.turn == TURN_BOARD_ACTION then
			do_move("auto")
		else
			local pid = current_turn()
			local m = select_move()
			do_move(m)
		end
	end
end

local load_ai_personality_prefix = ""
function load_ai_personality(name)
	filename = string.format("%s%s", load_ai_personality_prefix, name)

	local f = assert(loadfile(filename))

	local old_prefix = load_ai_personality_prefix
	local xx = string.match(filename, "^.*[/\\]")
	if xx then
		load_ai_personality_prefix = xx
	end

	local rv = f()

	load_ai_personality_prefix = old_prefix
	return rv
end

function load_ai_mixin(tmp_ai, name)
	local old_ai = ai
	ai = tmp_ai
	local rv = load_ai_personality(name)
	ai = old_ai
	return rv
end

function play_random_turn(personality)

	if state.turn == TURN_BOARD_ACTION then
		do_board_action()
		return "auto"
	else
		local m
		if personality then
			m = personality.select_move()
		else
			m = select_random_move()
		end
		do_move(m)
		return m
	end

end

function play_random_game_to_next_round(personality)

	local orig_round = state.round
	local principal_variation = {}
	while state.round == orig_round and state.castle_phase <= 3 do
		local m = play_random_turn(personality)
		table.insert(principal_variation, m)
	end

	return principal_variation
end

function play_random_game_to_next_scoring(personality)

	local orig_castle_phase = state.castle_phase
	if orig_castle_phase > 3 then return end

	while state.castle_phase == orig_castle_phase do
		play_random_game_to_next_round(personality)
	end

end

local start_time
function start_thinking()
	start_time = (get_tick_count or os.time)()
end

function finish_thinking()
	print(string.format("thought for %.2f seconds", get_elapsed()))
end

function get_elapsed()
	return (get_tick_count or os.time)() - start_time
end

function is_time_exceeded()
	return get_elapsed() >= 2.0
end
