local ai = load_ai_personality("commonguy.inc")

local function my_is_built(bldgref)
	if not ai.context.is_built_by_bref then
		local tmp = {}
		for i,bref in ipairs(WOOD_BUILDINGS) do
			tmp[bref] = is_built(bref)
		end
		for i,bref in ipairs(STONE_BUILDINGS) do
			tmp[bref] = is_built(bref)
		end
		for i,bref in ipairs(PRESTIGE_BUILDINGS) do
			tmp[bref] = is_built(bref)
		end
		ai.context.is_built_by_bref = tmp
	end
	return ai.context.is_built_by_bref[bldgref]
end
local is_built = my_is_built

local function can_build_building(bldgref, resources)
	if is_built(bldgref) then
		return false
	end

	if bldgref.cost == "1w,1_" then
		return resources.wood >= 1
			and
			(resources.food + resources.wood + resources.cloth + resources.stone - 1) >= 1
	end

	return check_cost_r(resources, bldgref.cost)
end

function ai.value_of_building(bldgref)
	return bldgref.btype == "wood" and 5 or 8
end

local function pick_payment(cost, r)
	if not r then
		local whoami = current_turn()
		local me = state.players[whoami]
		r = me.resources
	end

	if cost == "1w,1_" then
		if r.food >= 1 then
			return "1w,1f"
		elseif r.wood >= 2 then
			return "2w"
		elseif r.cloth >= 1 then
			return "1w,1c"
		elseif r.stone >= 1 then
			return "1w,1s"
		else
			error("can't pay")
		end
	end

	return cost
end

local function select_builder_move_helper(building_set)

	local whoami = current_turn()
	local me = state.players[whoami]

	local best_move = "noaction"
	local best_score = 0

	for i,bref in ipairs(building_set) do
		if can_build_building(bref, me.resources) then
			local my_score = ai.value_of_building(bref)
			if my_score > best_score then
				best_score = my_score
				best_move = string.format("build %s %s na", bref.id, pick_payment(bref.cost))
			end
		end
	end
	return best_move

end

function ai.select_builder_move()
	local bldgref = state.buildings[state.active_building].ref

	if bldgref == CARPENTER1 or bldgref == CARPENTER2 then
		--building a wooden building

		return select_builder_move_helper(WOOD_BUILDINGS)

	elseif bldgref == MASON then
		--building a stone building

		return select_builder_move_helper(STONE_BUILDINGS)

	elseif bldgref == LAWYER then

		return ai.select_lawyer_move()

	elseif bldgref == ARCHITECT1 or bldgref == ARCHITECT2 then

		return ai.select_architect_move()

	end

	--shouldn't reach here
	assert(false)
end

function ai.select_lawyer_move()
	local moves_array = all_available_moves()
	if #moves_array >= 2 and moves_array[1] == "noaction" then
		return moves_array[2]
	else
		return moves_array[1]
	end
end

function ai.select_architect_move()
	local moves_array = all_available_moves()
	if #moves_array >= 2 and moves_array[1] == "noaction" then
		return moves_array[2]
	else
		return moves_array[1]
	end
end

local function calculate_resources_for_unbuilt_buildings()
	if not ai.context.resources_for_unbuilt_buildings then
		local bag = resource_bag.new()
		local count = 0
		for i,bref in ipairs(WOOD_BUILDINGS) do
			if not is_built(bref) then
				if bref.cost == "1w,1c" then
					bag.wood = bag.wood + 1
					bag.cloth = bag.cloth + 1
				else
					bag.food = bag.food + 1
					bag.wood = bag.wood + 1
				end
				count = count + 1
			end
		end
		for i,bref in ipairs(STONE_BUILDINGS) do
			if not is_built(bref) then
				if bref.cost == "1s,1w" then
					bag.stone = bag.stone + 1
					bag.wood = bag.wood + 1
				elseif bref.cost == "1s,1c" then
					bag.stone = bag.stone + 1
					bag.cloth = bag.cloth + 1
				else
					bag.stone = bag.stone + 1
					bag.food = bag.food + 1
				end
				count = count + 1
			end
		end
		if ai.VERBOSE then
			print(string.format("want to get %s more resources for %d more buildings",
				resource_bag.as_string(bag),
				count))
		end
		bag.building_count = count
		bag.resource_count = bag.food + bag.wood + bag.stone + bag.cloth
		ai.context.resources_for_unbuilt_buildings = bag
	end
	return ai.context.resources_for_unbuilt_buildings
end

--@param r a resources struct
--
function ai.value_of_resources(r)
	--'a' is in [0,1], represents how far along in the game we are
	--we use this to prefer wood near the beginning of the game, but
	--stone towards the end of the game.
	local a = math.min(1, (state.round - 1) / 12)

	--value resources based on what buildings can still be built
	local good_for_bldgs = calculate_resources_for_unbuilt_buildings()
	local food_div = math.max(4, 4 + r.food - good_for_bldgs.food)
	local wood_div = math.max(4, 4 + r.wood - good_for_bldgs.wood)
	local stone_div = math.max(4, 4 + r.stone - good_for_bldgs.stone)

	local value_of_food
	do
		value_of_food = .25 * math.max(0, r.food - good_for_bldgs.food)
		local x = math.min(math.max(0,r.food), good_for_bldgs.food)
		if x > 0 then
			value_of_food = value_of_food
				+ 2.0 * x
				- (1.75 / 2) * x * x / good_for_bldgs.food
		end
	end
	local value_of_wood
	do
		value_of_wood = .25 * math.max(0, r.wood - good_for_bldgs.wood)
		local x = math.min(math.max(0,r.wood), good_for_bldgs.wood)
		if x > 0 then
			value_of_wood = value_of_wood
				+ 2.0 * x
				- (1.75 / 2) * x * x / good_for_bldgs.wood
		end
	end
	local value_of_stone
	do
		value_of_stone = .25 * math.max(0, r.stone - good_for_bldgs.stone)
		local x = math.min(math.max(0,r.stone), good_for_bldgs.stone)
		if x > 0 then
			value_of_stone = value_of_stone
				+ 2.0 * x
				- (1.75 / 2) * x * x / good_for_bldgs.stone
		end
	end

	return
		value_of_food +
		value_of_wood +
		value_of_stone +
		2.5 * r.cloth * 4 / (4 + r.cloth) +
		7.0 * r.gold  * 4 / (4 + r.gold) +
		2.5 * r.money * 4 / (4 + r.money)
end

function ai.select_worker_action()
	local bldgref = state.buildings[state.active_building].ref

	if state.turn == TURN_OWNER_ACTION then
		local s,m = ai:pick_building_owner_action(bldgref)
		return m
	end

	local s,m = ai:pick_building_pickup_action(bldgref)
	if s > 0 then
		return m
	end

	local moves_array = all_available_moves()
	if #moves_array >= 2 and moves_array[1] == "noaction" then
		return moves_array[2]
	else
		return moves_array[1]
	end
end

local function logisticfn(t)
	return 1 / (1 + math.exp(-t))
end

local function can_build_building(bldgref, resources)
	if is_built(bldgref) then
		return false
	end

	if bldgref.cost == "1w,1_" then
		return resources.wood >= 1
			and
			(resources.food + resources.wood + resources.cloth + resources.stone - 1) >= 1
	end

	return check_cost_r(resources, bldgref.cost)
end

--return an array of building refs
--
local function what_buildings_can_i_build(btype, resources)
	local building_set
	if btype == "wood" then
		building_set = WOOD_BUILDINGS
	elseif btype == "stone" then
		building_set = STONE_BUILDINGS
	else
		error("invalid btype: "..btype)
	end

	local result = {}
	for i,bldgref in ipairs(building_set) do
		if can_build_building(bldgref, resources) then
			table.insert(result, bldgref)
			if ai.VERBOSE then
				print(string.format("can build %s with %s",
					bldgref.id, resource_bag.as_string(resources)
					))
			end
		end
	end
	return result
end

local function count_blocks_already_given_this_phase()
	local whoami = current_turn()
	local x0 = state.castle_phase > 1
		and CASTLE_CONTRIBUTION_THRESHOLDS[state.castle_phase - 1] + 1
		or 1
	local x1 = CASTLE_CONTRIBUTION_THRESHOLDS[state.castle_phase]

	local count = 0
	for i = x0,x1 do
		if state.contributions[i] == whoami then
			count = count + 1
		end
	end
	return count
end

local function value_of_taking_castle(r, existing_occupants)

	--no point in taking castle if there's nothing to contribute
	if ai.estimate_maximum_contribution(r) < 1 then
		return -2
	end

	local whoami = current_turn()
	local i_would_be_first = existing_occupants[1] == whoami or not existing_occupants[1]
	local num_blocks_already_given_this_phase = count_blocks_already_given_this_phase()
	assert(state.castle_phase <= 3)
	local num_rounds_left_this_phase = (CASTLE_BAILIFF_THRESHOLDS[state.castle_phase] - state.bailiff) * .75

	if state.castle_phase == 3 and num_rounds_left_this_phase < 3 then
		return 8
	end

	if num_blocks_already_given_this_phase == 0 then
		if i_would_be_first and state.castle_phase > 1 then
			return 8
		elseif num_rounds_left_this_phase <= 2 then
			return 2 + 2 * state.castle_phase
		else
			return 1.5
		end
	elseif num_blocks_already_given_this_phase == 1 then
		if i_would_be_first then
			return 6
		else
			--having two blocks earns a favor,
			--but want to save resources in first phase of game
			if state.castle_phase < 3 then
				return 0
			else
				return 2
			end
		end
	else
		if i_would_be_first then
			return 2 --for the favor?
		else
			return 0
		end
	end
end

local function value_of_taking_builder(r, building_set)
	local best_score = -math.huge
	for i,bldgref in ipairs(building_set) do

		if can_build_building(bldgref, r) then

			local bag = resource_bag.parse(pick_payment(bldgref.cost, r))
			bag = resource_bag.scale(bag, -1)
			local cost_value = ai:value_of_resources_delta(bag, r)

			local my_score = ai.value_of_building(bldgref) + cost_value
			if my_score > best_score then
				best_score = my_score
			end
		end
	end

	return best_score
end

function ai.value_of_worker_at(i, r)
	local whoami = current_turn()
	local bldgref = assert(state.buildings[i].ref)

	local my_score = ai:pick_building_pickup_action(bldgref)
	if bldgref == CARPENTER1 or bldgref == CARPENTER2 then
		return value_of_taking_builder(r, WOOD_BUILDINGS)
	elseif bldgref == MASON then
		return value_of_taking_builder(r, STONE_BUILDINGS)
	elseif bldgref == STABLES then
		if #state.buildings[i].occupants == 0 then
			my_score = (position_in_turn_order(whoami) - 1) * 1.5
		end
	elseif bldgref == CASTLE then
		my_score = value_of_taking_castle(r, state.buildings[i].occupants)
	end
	if i < CASTLE_POSITION then
		my_score = my_score * logisticfn(state.provost - i)
	end
	return my_score
end

local function adjust_obtained(r, bldgref)
	if bldgref.actions then
		local s,movestr = ai:pick_building_pickup_action(bldgref)
		local rsrc_str = string.match(movestr, "^take (.*)$")
		if rsrc_str then
			local bag = resource_bag.parse(rsrc_str)
			r.food = r.food + bag.food
			r.wood = r.wood + bag.wood
			r.stone = r.stone + bag.stone
			r.cloth = r.cloth + bag.cloth
			r.gold = r.gold + bag.gold
			r.money = r.money + bag.money
			return
		end
	end

	if bldgref == CARPENTER1 or bldgref == CARPENTER2 then
		--note: for simplicity sake, we assume the
		--carpenter is going to cost us a wood+food
		r.wood = math.max(0, r.wood - 1)
		if r.food >= 1 then
			r.food = r.food - 1
		elseif r.cloth >= 1 then
			r.cloth = r.cloth - 1
		elseif r.stone >= 1 then
			r.stone = r.stone - 1
		else
			r.wood = math.max(0, r.wood - 1)
		end
	elseif bldgref == MASON then
		--note: for simplicity sake, we assume the
		--mason is going to cost us a stone+food
		r.stone = math.max(0, r.stone - 1)
		r.food = math.max(0, r.food - 1)
	elseif bldgref == LAWYER then
		r.money = r.money - 1
		r.cloth = r.cloth - 1
	elseif bldgref == BANK then
		if r.money >= 2 then
			r.money = r.money - 2
			r.gold = r.gold + 1
		end
	elseif bldgref == JOUST_FIELD then
		r.money = math.max(0, r.money - 1)
		r.cloth = math.max(0, r.cloth - 1)
	elseif bldgref == CASTLE then
		r.food = math.max(0, r.food - 1)
	end
end

local function in_building(whoami, bldg, start_index)
	local i = start_index or 1
	if bldg.occupants[i] == whoami then
		return true
	elseif bldg.occupants[i] then
		return in_building(whoami, bldg, i + 1)
	else
		return false
	end
end

local function calculate_existing_worker_values()
	local whoami = current_turn()
	local me = state.players[whoami]

	local r = resource_bag.clone(me.resources)
	local existing_worker_values = {}
	for i = 1, CASTLE_POSITION do
		if in_building(whoami, state.buildings[i]) then
			existing_worker_values[i] = ai.value_of_worker_at(i, r)
			adjust_obtained(r, state.buildings[i].ref)

		end
	end
	return existing_worker_values
end

--We've identified a move at 'i' that seems to be good,
--but we do not know yet whether it interferes with a
--worker we already placed further down the road.
--(E.g. we don't want to take Joust if we only have one
--cloth and we're already planning to spend the cloth
--on a Lawyer.)
--
local function check_for_interference(r, i, existing_worker_values)

	assert(existing_worker_values)

	local whoami = current_turn()
	local rr = resource_bag.clone(r)
	adjust_obtained(rr, state.buildings[i].ref)

	local sum = 0
	for j = i + 1, CASTLE_POSITION do
		if in_building(whoami, state.buildings[j]) then
			local delta = ai.value_of_worker_at(j, rr) - existing_worker_values[j]
			if delta ~= 0 and ai.VERBOSE then

				print(string.format("value of %s adjusted by %.1f from placing %s",
					state.buildings[j].ref.id,
					delta,
					state.buildings[i].ref.id))
			end
			sum = sum + delta
			adjust_obtained(rr, state.buildings[j].ref)
		end
	end
	return sum
end

function ai.select_worker_placement()
	local whoami = current_turn()
	local me = state.players[whoami]
	local valid_moves = {}
	for i,m in ipairs(all_available_moves()) do
		valid_moves[m] = true
	end

	--Precalculate the value of each worker already placed.
	--When a new worker is considered, we check the successor
	--workers to see if any of their values changed (as a
	--result of activating the newly placed worker).
	local existing_worker_values = calculate_existing_worker_values()

	--As we go down the road, we keep track of what resources we
	--will have at activation time.
	local r = resource_bag.clone(me.resources)

	local best_move = "pass"
	local best_score = 0
	for i = 1, CASTLE_POSITION do
		local bldgref = state.buildings[i].ref
		if state.buildings[i].occupants[1] == whoami then
			assert(bldgref)
			adjust_obtained(r, bldgref)
		end

		local movestr = string.format("claim %d", i)
		if valid_moves[movestr] then
			assert(bldgref)
			local my_score = ai.value_of_worker_at(i, r)
			if my_score > best_score then

				my_score = my_score + check_for_interference(r, i, existing_worker_values)
			end

			local bldg_cost
			do
				local bag = resource_bag.new()
				bag.money = -get_cost_to_place_worker(whoami, state.buildings[i])
				bldg_cost = ai:value_of_resources_delta(bag)
			end

			if ai.VERBOSE then
				print(string.format("move %s (%s) is worth %.1f (%.1f)",
					movestr,
					bldgref.id,
					my_score,
					bldg_cost))
			end

			my_score = my_score + bldg_cost
			if my_score > best_score then
				best_move = movestr
				best_score = my_score
			end

		end
	end

	if ai.VERBOSE then
		print("selected " .. best_move)
	end

	return best_move
end

function ai.select_favor_move()
	local whoami = current_turn()
	local me = state.players[whoami]

	local favor_options = {
		vp = {},
		money = {},
		resources = {},
		builder = {},
		builder_nil = {},
		}
	for i,m in ipairs(all_available_moves()) do
		if string.match(m, "^favor (%d+)vp$") then
			table.insert(favor_options.vp, m)
		elseif string.match(m, "^favor (%d+)%$$") then
			table.insert(favor_options.money, m)
		elseif m == "favor build" then
			table.insert(favor_options.builder_nil, m)
		elseif string.match(m, "^favor wood ")
			or string.match(m, "^favor stone ")
			or string.match(m, "^favor residence ")
			or string.match(m, "^favor prestige ")
		then
			table.insert(favor_options.builder, m)
		else
			table.insert(favor_options.resources, m)
		end
	end

	if me.resources.money < 3 and favor_options.money[1] then
		return favor_options.money[1]
	elseif favor_options.builder_nil[1] and me.build_favor == 0 then
		return favor_options.builder_nil[1]
	else
		return (all_available_moves())[1]
	end
end

return ai
