local ai = load_ai_personality("commonguy.inc")

function ai.value_of_building(bldgref)
	local bag = resource_bag.parse(bldgref.reward)
	return bag.score +
		4 * (bldgref.reward_favor or 0)
end

local function pick_payment(cost, r)
	if not r then
		local whoami = current_turn()
		local me = state.players[whoami]
		r = me.resources
	end

	if cost == "1w,1_" then
		if r.food >= 1 then
			return "1w,1f"
		elseif r.wood >= 2 then
			return "2w"
		elseif r.cloth >= 1 then
			return "1w,1c"
		elseif r.stone >= 1 then
			return "1w,1s"
		else
			error("can't pay")
		end
	end

	return cost
end

local function can_build_building(bldgref, resources)
	if is_built(bldgref) then
		return false
	end

	if bldgref.cost == "1w,1_" then
		return resources.wood >= 1
			and
			(resources.food + resources.wood + resources.cloth + resources.stone - 1) >= 1
	end

	return check_cost_r(resources, bldgref.cost)
end

local function select_builder_move_helper(building_set)

	local whoami = current_turn()
	local me = state.players[whoami]

	local best_move = "noaction"
	local best_score = 0

	for i,bref in ipairs(building_set) do
		if can_build_building(bref, me.resources) then
			local my_score = ai.value_of_building(bref)
			if my_score > best_score then
				best_score = my_score
				best_move = string.format("build %s %s na", bref.id, pick_payment(bref.cost))
			end
		end
	end
	return best_move

end

function ai.select_builder_move()
	local bldgref = state.buildings[state.active_building].ref

	if bldgref == CARPENTER1 or bldgref == CARPENTER2 then
		--building a wooden building

		return select_builder_move_helper(WOOD_BUILDINGS)

	elseif bldgref == MASON then
		--building a stone building

		return select_builder_move_helper(STONE_BUILDINGS)

	elseif bldgref == LAWYER then

		return ai.select_lawyer_move()

	elseif bldgref == ARCHITECT1 or bldgref == ARCHITECT2 then

		return ai.select_architect_move()

	end

	--shouldn't reach here
	assert(false)
end

function ai.select_lawyer_move()
	local moves_array = all_available_moves()
	if #moves_array >= 2 and moves_array[1] == "noaction" then
		return moves_array[2]
	else
		return moves_array[1]
	end
end

function ai.select_architect_move()
	local moves_array = all_available_moves()
	if #moves_array >= 2 and moves_array[1] == "noaction" then
		return moves_array[2]
	else
		return moves_array[1]
	end
end

--@param r a resources struct
--
function ai.value_of_resources(r)
	local base_value = 0
	local n_money = r.money
	local n_food = r.food
	local n_wood = r.wood
	local n_stone = r.stone
	local n_cloth = r.cloth

	if ai.context.position then
		local whoami = current_turn()
		local future_workers = {}

		for i = ai.context.position + 1, CASTLE_POSITION do
			local bldg = state.buildings[i]
			if bldg.ref and bldg.occupants[1] == whoami then
				future_workers[bldg.ref.id] = 1
			elseif bldg.ref and not bldg.occupants[1] then
				future_workers[bldg.ref.id] = .5
			end
		end

		if future_workers["church"] then
			if n_money >= 4 then
				base_value = base_value + 5
				n_money = n_money - 4
			elseif n_money >= 2 then
				base_value = base_value + 3
				n_money = n_money - 2
			end
		end
		if future_workers["castle"] and n_food >= 1 then
			if n_wood >= 1 and n_stone >= 1 then
				base_value = base_value + 5
				n_food = n_food - 1
				n_wood = n_wood - 1
				n_stone = n_stone - 1
			elseif n_wood >= 1 and n_cloth >= 1 then
				base_value = base_value + 5
				n_food = n_food - 1
				n_wood = n_wood - 1
				n_cloth = n_cloth - 1
			elseif n_stone >= 1 and n_cloth >= 1 then
				base_value = base_value + 5
				n_food = n_food - 1
				n_stone = n_stone - 1
				n_cloth = n_cloth - 1
			end
		end
	end

	base_value = base_value
		+ .333 * (n_food + n_wood + n_stone + n_cloth)
		+ 3 * r.gold
		+ .25 * n_money
	return base_value
end

local function pick_building_owner_action(bldgref)
	if bldgref.id == "inn" then
		return "noaction"
	end
	assert(bldgref.owner_actions and bldgref.owner_actions[1])

	local best_move
	local best_score = -math.huge

	for i,movestr in ipairs(bldgref.owner_actions) do
		
		local rsrc_str = string.match(movestr, "^take (.*)$")
		if rsrc_str then
			local bag = resource_bag.parse(rsrc_str)
			local my_score = ai:value_of_resources_delta(bag)
			if my_score > best_score then
				best_score = my_score
				best_move = movestr
			end
		end
	end

	assert(best_move)
	return best_score,best_move
end

local function pick_building_pickup_action(bldgref)

	if not (bldgref.actions and bldgref.actions[1]) then
		return 0,"noaction"
	end

	local best_move = "noaction"
	local best_score = 0

	for i,movestr in ipairs(bldgref.actions) do

		local rsrc_str = string.match(movestr, "^take (.*)$")
		if rsrc_str then
			local bag = resource_bag.parse(rsrc_str)
			local my_score = ai:value_of_resources_delta(bag)
			if my_score > best_score then
				best_score = my_score
				best_move = movestr
			end
		end
	end

	return best_score,best_move
end

function ai.select_worker_action()
	local bldgref = state.buildings[state.active_building].ref

	if state.turn == TURN_OWNER_ACTION then
		local s,m = pick_building_owner_action(bldgref)
		return m
	end

	local s,m = pick_building_pickup_action(bldgref)
	if s > 0 then
		return m
	end

	local moves_array = all_available_moves()
	local best_move
	local best_score = -math.huge
	for i,m in ipairs(moves_array) do
		local my_score = 0
		if m == "noaction" then
			my_score = -1
		end
		local exchange1,exchange2 = string.match(m, "^exchange (%S+) (%S+)$")
		if exchange1 and exchange2 then
			local bag1 = resource_bag.parse(exchange1)
			local bag2 = resource_bag.parse(exchange2)
			local bag = resource_bag.subtract(bag2, bag1)
			my_score = bag.score + ai:value_of_resources_delta(bag)
		end

		if my_score > best_score then
			best_move = m
			best_score = my_score
		end
	end
	return assert(best_move)
end

local function logisticfn(t)
	return 1 / (1 + math.exp(-t))
end

--return an array of building refs
--
local function what_buildings_can_i_build(btype, resources)
	local building_set
	if btype == "wood" then
		building_set = WOOD_BUILDINGS
	elseif btype == "stone" then
		building_set = STONE_BUILDINGS
	else
		error("invalid btype: "..btype)
	end

	local result = {}
	for i,bldgref in ipairs(building_set) do
		if can_build_building(bldgref, resources) then
			table.insert(result, bldgref)
			if ai.VERBOSE then
				print(string.format("can build %s with %s",
					bldgref.id, resource_bag.as_string(resources)
					))
			end
		end
	end
	return result
end

local function count_blocks_already_given_this_phase()
	local whoami = current_turn()
	local x0 = state.castle_phase > 1
		and CASTLE_CONTRIBUTION_THRESHOLDS[state.castle_phase - 1] + 1
		or 1
	local x1 = CASTLE_CONTRIBUTION_THRESHOLDS[state.castle_phase]

	local count = 0
	for i = x0,x1 do
		if state.contributions[i] == whoami then
			count = count + 1
		end
	end
	return count
end

local function value_of_taking_castle(r, existing_occupants)

	--no point in taking castle if there's nothing to contribute
	local n_able = ai.estimate_maximum_contribution(r)
	if n_able < 1 then
		return -2
	end

	local whoami = current_turn()
	local i_would_be_first = existing_occupants[1] == whoami or not existing_occupants[1]

	if i_would_be_first then
		return n_able * 3 + 4
	else
		return n_able * 3
	end
end

local function value_of_taking_builder(r, building_set)
	local best_score = -math.huge
	for i,bldgref in ipairs(building_set) do

		if can_build_building(bldgref, r) then

			local bag = resource_bag.parse(pick_payment(bldgref.cost, r))
			bag = resource_bag.scale(bag, -1)
			local cost_value = ai:value_of_resources_delta(bag,r)

			local my_score = ai.value_of_building(bldgref) + cost_value
			if my_score > best_score then
				best_score = my_score
			end
		end
	end

	return best_score
end

local function value_of_taking_alchemist(r)
	local n_resources = r.food + r.wood + r.stone + r.cloth
	if n_resources >= 4 then
		return 6 - 4/3
	elseif n_resources >= 2 then
		return 3 - 2/3
	else
		return 0
	end
end

local function value_of_taking_bank(r)
	local bag = resource_bag.new()
	if r.money >= 5 then
		bag.money = -5
		bag.gold = 2
		return ai:value_of_resources_delta(bag,r)
	elseif r.money >= 2 then
		bag.money = -2
		bag.gold = 1
		return ai:value_of_resources_delta(bag,r)
	else
		return 0
	end
end

local function value_of_taking_church(r)
	local bag = resource_bag.new()
	if r.money >= 4 then
		bag.money = -4
		return 5 + ai:value_of_resources_delta(bag,r)
	elseif r.money >= 2 then
		bag.money = -2
		return 3 + ai:value_of_resources_delta(bag,r)
	else
		return 0
	end
end

local function value_of_taking_jeweler(r)
	local bag = resource_bag.new()
	if r.gold >= 2 then
		bag.gold = -2
		return 9 + ai:value_of_resources_delta(bag,r)
	elseif r.gold >= 1 then
		bag.gold = -1
		return 5 + ai:value_of_resources_delta(bag,r)
	else
		return 0
	end
end

local function value_of_taking_tailor(r)
	local bag = resource_bag.new()
	if r.cloth >= 3 then
		bag.cloth = -3
		return 6 + ai:value_of_resources_delta(bag,r)
	elseif r.cloth >= 2 then
		bag.cloth = -2
		return 4 + ai:value_of_resources_delta(bag,r)
	else
		return 0
	end
end

local function pick_marketplace_action(r, bldgref)
	local best_move = "noaction"
	local best_score = 0
	for i,resource in ipairs({"food","wood","stone","cloth"}) do
		if r[resource] >= 1 then
			local give_bag = resource_bag.new()
			give_bag[resource] = 1
			local take_bag = resource_bag.new()
			take_bag.money = (bldgref == MARKETPLACE1 and 4 or 6)
			local my_score = ai:value_of_resources_delta(resource_bag.subtract(take_bag, give_bag))
			if my_score > best_score then
				best_score = my_score
				best_move = string.format("exchange %s %s",
					resource_bag.as_string(give_bag),
					resource_bag.as_string(take_bag))
			end
		end
	end
	return best_score,best_move
end

local function value_of_taking_marketplace(r, bldgref)
	local s = pick_marketplace_action(r, bldgref)
	return s
end

function ai.value_of_worker_at(i, r)
	local whoami = current_turn()
	local bldgref = assert(state.buildings[i].ref)

	ai.context.position = i
	local my_score = pick_building_pickup_action(bldgref)
	if bldgref == CARPENTER1 or bldgref == CARPENTER2 then
		return value_of_taking_builder(r, WOOD_BUILDINGS)
	elseif bldgref == MASON then
		return value_of_taking_builder(r, STONE_BUILDINGS)
	elseif bldgref == STABLES then
		return 0
	elseif bldgref == MARKETPLACE1 or bldgref == MARKETPLACE2 then
		return value_of_taking_marketplace(r, bldgref)
	elseif bldgref == ALCHEMIST then
		return value_of_taking_alchemist(r)
	elseif bldgref == BANK then
		return value_of_taking_bank(r)
	elseif bldgref == CHURCH then
		return value_of_taking_church(r)
	elseif bldgref == JEWELER then
		return value_of_taking_jeweler(r)
	elseif bldgref == TAILOR then
		return value_of_taking_tailor(r)
	elseif bldgref == CASTLE then
		my_score = value_of_taking_castle(r, state.buildings[i].occupants)
	end
	if i < CASTLE_POSITION then
		my_score = my_score * logisticfn(state.provost - i)
	end
	return my_score
end

local function adjust_obtained(r, bldgref)
	if bldgref.actions then
		local s,movestr = pick_building_pickup_action(bldgref)
		local rsrc_str = string.match(movestr, "^take (.*)$")
		if rsrc_str then
			local bag = resource_bag.parse(rsrc_str)
			r.food = r.food + bag.food
			r.wood = r.wood + bag.wood
			r.stone = r.stone + bag.stone
			r.cloth = r.cloth + bag.cloth
			r.gold = r.gold + bag.gold
			r.money = r.money + bag.money
			return
		end
	end

	if bldgref == CARPENTER1 or bldgref == CARPENTER2 then
		--note: for simplicity sake, we assume the
		--carpenter is going to cost us a wood+food
		r.wood = math.max(0, r.wood - 1)
		if r.food >= 1 then
			r.food = r.food - 1
		elseif r.cloth >= 1 then
			r.cloth = r.cloth - 1
		elseif r.stone >= 1 then
			r.stone = r.stone - 1
		else
			r.wood = math.max(0, r.wood - 1)
		end
	elseif bldgref == MASON then
		--note: for simplicity sake, we assume the
		--mason is going to cost us a stone+food
		r.stone = math.max(0, r.stone - 1)
		r.food = math.max(0, r.food - 1)
	elseif bldgref == LAWYER then
		r.money = r.money - 1
		r.cloth = r.cloth - 1
	elseif bldgref == BANK then
		if r.money >= 2 then
			r.money = r.money - 2
			r.gold = r.gold + 1
		end
	elseif bldgref == JOUST_FIELD then
		r.money = math.max(0, r.money - 1)
		r.cloth = math.max(0, r.cloth - 1)
	elseif bldgref == CASTLE then
		r.food = math.max(0, r.food - 1)
	end
end

local function in_building(whoami, bldg, start_index)
	local i = start_index or 1
	if bldg.occupants[i] == whoami then
		return true
	elseif bldg.occupants[i] then
		return in_building(whoami, bldg, i + 1)
	else
		return false
	end
end

local function calculate_existing_worker_values()
	local whoami = current_turn()
	local me = state.players[whoami]

	local r = resource_bag.clone(me.resources)
	local existing_worker_values = {}
	for i = 1, CASTLE_POSITION do
		if in_building(whoami, state.buildings[i]) then
			existing_worker_values[i] = ai.value_of_worker_at(i, r)
			adjust_obtained(r, state.buildings[i].ref)

		end
	end
	return existing_worker_values
end

--We've identified a move at 'i' that seems to be good,
--but we do not know yet whether it interferes with a
--worker we already placed further down the road.
--(E.g. we don't want to take Joust if we only have one
--cloth and we're already planning to spend the cloth
--on a Lawyer.)
--
local function check_for_interference(r, i, existing_worker_values)

	assert(existing_worker_values)

	local whoami = current_turn()
	local rr = resource_bag.clone(r)
	adjust_obtained(rr, state.buildings[i].ref)

	local sum = 0
	for j = i + 1, CASTLE_POSITION do
		if in_building(whoami, state.buildings[j]) then
			local delta = ai.value_of_worker_at(j, rr) - existing_worker_values[j]
			if delta ~= 0 and ai.VERBOSE then

				print(string.format("value of %s adjusted by %.1f from placing %s",
					state.buildings[j].ref.id,
					delta,
					state.buildings[i].ref.id))
			end
			sum = sum + delta
			adjust_obtained(rr, state.buildings[j].ref)
		end
	end
	return sum
end

function ai.select_worker_placement()
	local whoami = current_turn()
	local me = state.players[whoami]
	local valid_moves = {}
	for i,m in ipairs(all_available_moves()) do
		valid_moves[m] = true
	end

	--Precalculate the value of each worker already placed.
	--When a new worker is considered, we check the successor
	--workers to see if any of their values changed (as a
	--result of activating the newly placed worker).
	local existing_worker_values = calculate_existing_worker_values()

	--As we go down the road, we keep track of what resources we
	--will have at activation time.
	local r = resource_bag.clone(me.resources)

	local best_move = "pass"
	local best_score = 0
	for i = 1, CASTLE_POSITION do
		local bldgref = state.buildings[i].ref
		if state.buildings[i].occupants[1] == whoami then
			assert(bldgref)
			adjust_obtained(r, bldgref)
		end

		local movestr = string.format("claim %d", i)
		if valid_moves[movestr] then
			assert(bldgref)
			local my_score = ai.value_of_worker_at(i, r)
			if my_score > best_score then

				my_score = my_score + check_for_interference(r, i, existing_worker_values)
			end

			local bldg_cost
			do
				local bag = resource_bag.new()
				bag.money = -get_cost_to_place_worker(whoami, state.buildings[i])
				bldg_cost = ai:value_of_resources_delta(bag)
			end

			if ai.VERBOSE then
				print(string.format("move %s (%s) is worth %.1f (%.1f)",
					movestr,
					bldgref.id,
					my_score,
					bldg_cost))
			end

			my_score = my_score + bldg_cost
			if my_score > best_score then
				best_move = movestr
				best_score = my_score
			end

		end
	end

	if ai.VERBOSE then
		print("selected " .. best_move)
	end

	return best_move
end

function ai.select_favor_move()
	local whoami = current_turn()
	local me = state.players[whoami]

	local best_move
	local best_score = -math.huge
	for i,m in ipairs(all_available_moves()) do
		local my_score
		if string.match(m, "^favor (%d+)vp$") then
			my_score = tonumber(string.match(m, "^favor (%d+)vp$"))
		elseif string.match(m, "^favor (%d+)%$$") then
			local tmp = tonumber(string.match(m, "^favor (%d+)%$$"))
			local new_money = me.resources.money + tmp
			my_score = math.floor(new_money/4) - math.floor(me.resources.money/4)
		elseif m == "favor build" then
			my_score = 0
		elseif string.match(m, "^favor wood ")
			or string.match(m, "^favor stone ")
			or string.match(m, "^favor residence ")
			or string.match(m, "^favor prestige ")
		then
			my_score = 3
		elseif m == "favor 1g" then
			my_score = 3
		else
			--simple resource
			local n_resources = me.resources.food + me.resources.wood
				+ me.resources.stone
				+ me.resources.cloth
			my_score = math.floor((n_resources+1)/3)-math.floor(n_resources/3)
		end

		if my_score > best_score then
			best_score = my_score
			best_move = m
		end
	end

	return best_move
end

--load_ai_mixin(ai, "resource_appraiser2.inc")
return ai
