local WANT_MONEY_THRESHOLD = 6
local VALUE_OF_CARPENTER = 2
local ai = {}
local plan = {
		on_carpenter = false, --true if worker placed on a carpenter
		on_castle = false,    --true if worker placed on castle
		resources_at_castle = {}, --what resources expect to have at castle
		gives = {             --for each resource, a list of positions that give that resource
			food = {}, wood = {}, stone = {}, cloth = {}, gold = {},
			},
		expected_result = {}, --what resources expect to have after buildings with me on them
	}

local function get_payout_from_action(action)
	if action == "take 2s,1c" then
		return { 0,0,2,1 }
	elseif action == "take 2f,1c" then
		return { 2,0,0,1 }
	elseif action == "take 2w,1f" then
		return { 1,2,0,0 }
	elseif action == "take 2f" then
		return { 2,0,0,0 }
	elseif action == "take 2w" then
		return { 0,2,0,0 }
	elseif action == "take 2s" then
		return { 0,0,2,0 }
	elseif action == "take 2c" then
		return { 0,0,0,2 }
	elseif action == "take 1f" then
		return { 1,0,0,0 }
	elseif action == "take 1w" then
		return { 0,1,0,0 }
	elseif action == "take 1s" then
		return { 0,0,1,0 }
	elseif action == "take 1c" then
		return { 0,0,0,1 }
	elseif action == "take 1g" then
		return { 0,0,0,0,1 }
	else
		return nil
	end
end

local function estimate_maximum_contribution(r)
	return math.max(
			math.min(r.food, r.wood, r.stone),
			math.min(r.food, r.wood, r.cloth),
			math.min(r.food, r.stone, r.cloth)
			)
end

local function evaluate_pickup_action(r, action)
	assert(r.food >= 0, "negative food")
	assert(r.wood >= 0, "negative food")
	assert(r.stone >= 0, "negative food")
	assert(r.cloth >= 0, "negative food")
	local payout = get_payout_from_action(action)
	if payout then
		return
			payout[1] / (r.food + .9) +
			payout[2] / (r.wood + 1) +
			payout[3] / (r.stone + 1) +
			payout[4] / (r.cloth + 1)
	else
		return 0
	end
end

local function choose_pickup_action(r, bldgref)
	assert(bldgref.actions)
	local best_move
	local best_move_value = -1
	for j,action in ipairs(bldgref.actions) do
		local this_move_value = evaluate_pickup_action(r, action)
		if this_move_value > best_move_value then
			best_move_value = this_move_value
			best_move = action
		end
	end --next action at this building
	assert(best_move, "no best move for building "..bldgref.name)
	return best_move
end

local function build_plan(me)
	local plan = {
		gives = { food = {}, wood = {}, stone = {}, cloth = {}, },
		expected_result = {},
		}

	local r = {
		food = me.resources.food,
		wood = me.resources.wood,
		stone = me.resources.stone,
		cloth = me.resources.cloth,
		gold = me.resources.gold,
		}

	local function i_am_here(position)
		for i,pid in ipairs(state.buildings[position].occupants) do
			if state.players[pid] == me then
				return true
			end
		end
		return false
	end
	local function is_occupied(position)
		return #state.buildings[position].occupants > 0
	end

	--find the workers I've already placed
	for i = BRIDGE_POSITION+1, state.bailiff do
		if i_am_here(i) then
			local bldgref = state.buildings[i].ref
			if bldgref == CARPENTER1 or bldgref == CARPENTER2 then
				r.food = math.max(0, r.food - 1)
				r.wood = math.max(0, r.wood - 1)
				plan.on_carpenter = true
			elseif bldgref.actions then
				local action = choose_pickup_action(r, bldgref)
				local payout = get_payout_from_action(action)
				assert(payout, "no payout for "..action)
				r.food = r.food + payout[1]
				r.wood = r.wood + payout[2]
				r.stone = r.stone + payout[3]
				r.cloth = r.cloth + payout[4]
			end
			plan.expected_result[i] = {
					food = r.food,
					wood = r.wood,
					stone = r.stone,
					cloth = r.cloth,
					gold = r.gold,
				}
		elseif not is_occupied(i) then
			local bldgref = state.buildings[i].ref
			if bldgref and bldgref.actions then
				local can_get = {}
				for j,action in ipairs(bldgref.actions) do
					local payout = get_payout_from_action(action)
					if payout then
						if payout[1] > 0 then can_get.food = 1 end
						if payout[2] > 0 then can_get.wood = 1 end
						if payout[3] > 0 then can_get.stone = 1 end
						if payout[4] > 0 then can_get.cloth = 1 end
					end
				end
				for rsrc,x in pairs(can_get) do
					table.insert(plan.gives[rsrc], i)
				end
			end
		end
	end --next building
	if i_am_here(CASTLE_POSITION) then
		plan.on_castle = true
	end

	plan.resources_at_castle = r
	return plan
end

local function evaluate_worker_placement_choice(me, bldgref)
	local r = plan.resources_at_castle
	if bldgref == CARPENTER1 or bldgref == CARPENTER2 then
		local any_wood_production_building_left
			= not is_built(WOOD_BUILDINGS[1]) 
				or not is_built(WOOD_BUILDINGS[2])
				or not is_built(WOOD_BUILDINGS[3])
				or not is_built(WOOD_BUILDINGS[4])
		if not plan.on_carpenter then
			if any_wood_production_building_left then
				local v = VALUE_OF_CARPENTER
				if state.buildings[state.bailiff + 1].ref
					and state.buildings[state.bailiff + 2].ref
					then
					--any building built now won't be available until a
					--couple rounds in the future, so it's not so useful
					--to build now
					v = v / 2
				end
				if plan.on_castle then
					if estimate_maximum_contribution(r) >= 1 and r.food >= 2 and r.wood >= 2 then
						return v
					end
				else
					if r.food >= 1 and r.wood >= 1 then
						return v
					end
				end
			end
		end
		return -1
	elseif bldgref.actions then
				--is this a building I want?
				local best_value = 0
				for j,action in ipairs(bldgref.actions) do
					local this_move_value = evaluate_pickup_action(r, action)
					if this_move_value > best_value then
						best_value = this_move_value
					end
				end --next action at this building
				return best_value
	end --has actions

	--do not want to place a worker here
	return -1
end

function ai.select_move()
	local whoami = current_turn()
	local me = state.players[whoami]
	local r = me.resources

	plan = build_plan(me)

	local ph = current_phase()
	if ph == PHASE_WORKER_PLACEMENT then
		if me.resources.money < 1 then
			--automatic
			return "pass"
		end
		r = plan.resources_at_castle
		local estimated_contribution = estimate_maximum_contribution(r)
		if estimated_contribution >= 1
			and #state.buildings[CASTLE_POSITION].occupants == 0
			and me.resources.money >= get_cost_to_place_worker(whoami, state.buildings[CASTLE_POSITION])
			then
			return string.format("claim %d",CASTLE_POSITION)
		end
		if estimated_contribution >= 3
			and not plan.on_castle
			and me.resources.money >= get_cost_to_place_worker(whoami, state.buildings[CASTLE_POSITION])
			then
			return string.format("claim %d", CASTLE_POSITION)
		end
		if state.bailiff >= GOLDMINE_POSITION
			and #state.buildings[GOLDMINE_POSITION].occupants == 0
			and me.resources.money >= get_cost_to_place_worker(whoami, state.buildings[GOLDMINE_POSITION])
			then
			return string.format("claim %d", GOLDMINE_POSITION)
		end
		if me.resources.money < WANT_MONEY_THRESHOLD
			and #state.buildings[TRADING_POST_POSITION].occupants == 0
			and me.resources.money >= get_cost_to_place_worker(whoami, state.buildings[TRADING_POST_POSITION])
			then
			return string.format("claim %d", TRADING_POST_POSITION)
		end
		local best_move = "pass"
		local best_move_value = 0
		for i = BRIDGE_POSITION+1, state.bailiff do
			local bldgref = state.buildings[i].ref
			if bldgref and #state.buildings[i].occupants == 0
				and me.resources.money >= get_cost_to_place_worker(whoami, state.buildings[i])
				then
				local this_move_value = evaluate_worker_placement_choice(me, bldgref)
				if this_move_value > best_move_value then
					best_move = string.format("claim %d", i)
					best_move_value = this_move_value
				end
			end --if able to place worker
		end --next building
		return best_move
	elseif ph == PHASE_PICKUP_PROFIT then
		local bldgref = state.buildings[state.active_building].ref
		assert(bldgref.actions)
		return choose_pickup_action(r, bldgref)
	elseif ph == PHASE_CASTLE_BUILD then
		--don't want to use gold to contribute to castle
		local tg = r.gold
		r.gold = 0
		local mvs = available_moves_for_castle()
		r.gold = tg
		--any other players in castle after me?
		local more_players = #state.buildings[CASTLE_POSITION].occupants > state.turn
		local best_contribution_so_far = state.best_contribution_amount
		local function size_of_contribution(movestr)
			if movestr == "castle nobuild" then
				return 0
			else
				return math.floor((string.len(movestr) - string.len("castle ")) / 3)
			end
		end
		local function contribute_at_most(maximum)
			for i,movestr in ipairs(mvs) do
				if size_of_contribution(movestr) <= maximum then
					return movestr
				end
			end
		end
		local my_best_contribution = size_of_contribution(mvs[1])
		if not more_players and state.castle_phase <= 2 then
			--only contribute the minimum needed to earn the favor
			if my_best_contribution > best_contribution_so_far then
				return contribute_at_most(best_contribution_so_far + 1)
			else
				return contribute_at_most(1)
			end
		end
		--contribute maximum possible
		return mvs[1]
	elseif ph == PHASE_SELECT_BUILDING then
		if r.wood >= 1 and r.food >= 1 then
			if not is_built(WOOD_BUILDINGS[1]) then
				return string.format("build wood 1 1w,1f")
			elseif not is_built(WOOD_BUILDINGS[2]) then
				return string.format("build wood 2 1w,1f")
			elseif not is_built(WOOD_BUILDINGS[3]) then
				return string.format("build wood 3 1w,1f")
			elseif not is_built(WOOD_BUILDINGS[4]) then
				return string.format("build wood 4 1w,1f")
			end
		end
	else
		--assert(false, "current phase is "..ph)
	end
	return ai2_select()
end

return ai
