--Change log:
--  2010-01-16: created planner4; implement logic for provost movement;
--      allow choosing buildings beyond bailiff position;
--      further plan is to sometimes pick merchant guild

local TUNABLE = {
		WANT_MONEY_THRESHOLD = 6,   --[0,inf]
		VALUE_OF_CARPENTER = 2,     --[0,inf]
		VALUE_OF_WOODEN_PRODUCER = 2,
		VALUE_OF_MASON = 2,
		VALUE_OF_STONE_PRODUCER = 3,
	}
local WANT_MONEY_THRESHOLD = TUNABLE.WANT_MONEY_THRESHOLD
local ai = {}
local plan = {
		on_merchant_guild = false,
		on_carpenter = false, --true if worker placed on a carpenter
		on_castle = false,    --true if worker placed on castle
		resources_at_castle = {}, --what resources expect to have at castle
		gives = {             --for each resource, a list of positions that give that resource
			food = {}, wood = {}, stone = {}, cloth = {}, gold = {},
			},
		expected_result = {}, --what resources expect to have after buildings with me on them
		expected_provost = 0, --where we expect the provost to be
		last_of_my_workers = 0, --position of last of my workers
		last_of_opponent_workers = 0, --likewise for opponents' workers
	}

local function get_payout_from_action(action)
	if action == "take 2s,1c" then
		return { 0,0,2,1,0 }
	elseif action == "take 2f,1c" then
		return { 2,0,0,1,0 }
	elseif action == "take 2w,1f" then
		return { 1,2,0,0,0 }
	elseif action == "take 2f" then
		return { 2,0,0,0,0 }
	elseif action == "take 2w" then
		return { 0,2,0,0,0 }
	elseif action == "take 2s" then
		return { 0,0,2,0,0 }
	elseif action == "take 2c" then
		return { 0,0,0,2,0 }
	elseif action == "take 1f" then
		return { 1,0,0,0,0 }
	elseif action == "take 1w" then
		return { 0,1,0,0,0 }
	elseif action == "take 1s" then
		return { 0,0,1,0,0 }
	elseif action == "take 1c" then
		return { 0,0,0,1,0 }
	elseif action == "take 1g" then
		return { 0,0,0,0,1 }
	elseif action == "take 3$" then
		return { 0,0,0,0,0 }
	else
		return nil
	end
end

local function estimate_maximum_contribution(r)
	return math.max(
			math.min(r.food, r.wood, r.stone),
			math.min(r.food, r.wood, r.cloth),
			math.min(r.food, r.stone, r.cloth)
			)
end

local function evaluate_pickup_action(r, action)
	assert(r.food >= 0, "negative food")
	assert(r.wood >= 0, "negative food")
	assert(r.stone >= 0, "negative food")
	assert(r.cloth >= 0, "negative food")
	local payout = get_payout_from_action(action)
	if payout then
		return
			payout[1] / (r.food + .9) +
			payout[2] / (r.wood + 1) +
			payout[3] / (r.stone + 1) +
			payout[4] / (r.cloth + 1) +
			payout[5] * 3
	else
		return 0
	end
end

local function choose_pickup_action(r, bldgref)
	assert(bldgref.actions)
	local best_move
	local best_move_value = -1
	local available_actions = bldgref.actions
	if bldgref.owner_actions and state.turn == TURN_OWNER_ACTION then
		available_actions = bldgref.owner_actions
	end

	for j,action in ipairs(available_actions) do
		local this_move_value = evaluate_pickup_action(r, action)
		if this_move_value > best_move_value then
			best_move_value = this_move_value
			best_move = action
		end
	end --next action at this building
	assert(best_move, "no best move for building "..bldgref.name)
	return best_move
end

local function build_plan(me)
	local plan = {
		gives = { food = {}, wood = {}, stone = {}, cloth = {}, },
		expected_result = {},
		last_of_my_workers = 0,
		last_of_opponent_workers = 0,
		}

	local r = {
		food = me.resources.food,
		wood = me.resources.wood,
		stone = me.resources.stone,
		cloth = me.resources.cloth,
		gold = me.resources.gold,
		}

	local function i_am_here(position)
		for i,pid in ipairs(state.buildings[position].occupants) do
			if state.players[pid] == me then
				return true
			end
		end
		return false
	end
	local function is_occupied(position)
		return #state.buildings[position].occupants > 0
	end

	--find the workers I've already placed
	for i = 1, LAST_EMPTY_POSITION do
		if i_am_here(i) then
			plan.last_of_my_workers = i
			local bldgref = state.buildings[i].ref
			if bldgref == MERCHANT_GUILD then
				plan.on_merchant_guild = true
			elseif bldgref == CARPENTER1 or bldgref == CARPENTER2 then
				r.food = math.max(0, r.food - 1)
				r.wood = math.max(0, r.wood - 1)
				plan.on_carpenter = true
			elseif bldgref == MASON then
				r.food = math.max(0, r.food - 1)
				r.stone = math.max(0, r.stone - 1)
				plan.on_carpenter = true
			elseif bldgref.actions then
				local action = choose_pickup_action(r, bldgref)
				local payout = get_payout_from_action(action)
				assert(payout, "no payout for "..action)
				if i + 1 < state.provost then
					r.food = r.food + payout[1]
					r.wood = r.wood + payout[2]
					r.stone = r.stone + payout[3]
					r.cloth = r.cloth + payout[4]
				end
			end
			plan.expected_result[i] = {
					food = r.food,
					wood = r.wood,
					stone = r.stone,
					cloth = r.cloth,
					gold = r.gold,
				}
		elseif is_occupied(i) then
			--building is occupied by an opponent
			plan.last_of_opponent_workers = i
		else --building is empty
			local bldgref = state.buildings[i].ref
			if bldgref and bldgref.actions then
				local can_get = {}
				for j,action in ipairs(bldgref.actions) do
					local payout = get_payout_from_action(action)
					if payout then
						if payout[1] > 0 then can_get.food = 1 end
						if payout[2] > 0 then can_get.wood = 1 end
						if payout[3] > 0 then can_get.stone = 1 end
						if payout[4] > 0 then can_get.cloth = 1 end
					end
				end
				for rsrc,x in pairs(can_get) do
					table.insert(plan.gives[rsrc], i)
				end
			end
		end
	end --next building
	if i_am_here(CASTLE_POSITION) then
		plan.on_castle = true
	end

	plan.expected_provost = state.bailiff
	if plan.on_merchant_guild then
		if plan.last_of_my_workers > plan.last_of_opponent_workers
				or plan.last_of_opponent_workers < state.bailiff then
			plan.expected_provost = plan.expected_provost + 2
		else
			plan.expected_provost = plan.expected_provost - 2
		end
	end

	plan.resources_at_castle = r
	return plan
end

local function estimate_num_rounds_left()
	local bailiff_spaces_left = CASTLE_BAILIFF_THRESHOLDS[3] - state.bailiff
	if state.castle_phase < 3 then
		--TODO make this a tunable
		return math.floor(bailiff_spaces_left * 3/4)
	end

	local contribution_left = CASTLE_CONTRIBUTION_THRESHOLDS[3] - state.contribution_count
	return math.floor( math.min(
			bailiff_spaces_left,
			--TODO make this a tunable
			contribution_left * 6 / 14
			))
end

local function estimate_value_of_building(bldgref)
	--estimate how many rounds are left
	local rounds_left = estimate_num_rounds_left()

	--estimate how many rounds it will be before this building can be used
	local rounds_wait = rounds_left
	for i = state.bailiff, LAST_EMPTY_POSITION do
		if state.buildings[i].ref == nil then
			rounds_wait = i - state.bailiff
			break
		end
	end

	local factor = (rounds_left - rounds_wait) / 12

	if bldgref == WOOD_BUILDINGS[1]
		or bldgref == WOOD_BUILDINGS[2]
		or bldgref == WOOD_BUILDINGS[3]
		or bldgref == WOOD_BUILDINGS[4]
		then
		return factor * TUNABLE.VALUE_OF_WOODEN_PRODUCER
	elseif bldgref == MASON then
		return factor * TUNABLE.VALUE_OF_MASON
	elseif bldgref == STONE_BUILDINGS[1]
		or bldgref == STONE_BUILDINGS[2]
		or bldgref == STONE_BUILDINGS[3]
		then
		return factor * TUNABLE.VALUE_OF_STONE_PRODUCER
	else
		return 0
	end
end

local function first_empty_position()
	for i = FIRST_EMPTY_POSITION,LAST_EMPTY_POSITION do
		if not state.buildings[i].ref then
			return i
		end
	end
	return LAST_EMPTY_POSITION
end

local function evaluate_worker_placement_choice(me, position)
	local bldgref = state.buildings[position].ref
	local r = plan.resources_at_castle
	if bldgref == MERCHANT_GUILD then
		if state.bailiff >= first_empty_position() then
			return 0
		else
			return 1
		end
	elseif bldgref == CARPENTER1 or bldgref == CARPENTER2 then
		if plan.on_carpenter then
			--never attempt claiming multiple builders; too confusing
			return 0
		end

		local candidate_buildings = WOOD_BUILDINGS
		local best_move_value = 0
		for i,bldgref in ipairs(candidate_buildings) do
			if not is_built(bldgref) then
				--can we afford this building?
				local cost_str = bldgref.cost
				if cost_str == "1w,1_" then
					cost_str = "1w,1f"
				end
				local cost = resource_bag.parse(cost_str)
				if r.food >= cost.food
					and r.wood >= cost.wood
					and r.stone >= cost.stone
					and r.cloth >= cost.cloth
					and r.gold >= cost.gold
					then
					--able to afford this building
					local this_move_value = estimate_value_of_building(bldgref)
					if this_move_value > best_move_value then
						best_move_value = this_move_value
					end
				end --if affordable
			end --if not already built
		end --for each candidate building

		--TODO- if planning to use castle, ...
		return best_move_value
	elseif bldgref == MASON then
		if plan.on_carpenter then
			--never attempt claiming multiple builders; too confusing
			return 0
		end

		local candidate_buildings = STONE_BUILDINGS
		local best_move_value = 0
		for i,bldgref in ipairs(candidate_buildings) do
			if not is_built(bldgref) then
				--can we afford this building?
				local cost_str = bldgref.cost
				local cost = resource_bag.parse(cost_str)
				if r.food >= cost.food
					and r.wood >= cost.wood
					and r.stone >= cost.stone
					and r.cloth >= cost.cloth
					and r.gold >= cost.gold
					then
					--able to afford this building
					local this_move_value = estimate_value_of_building(bldgref)
					if this_move_value > best_move_value then
						best_move_value = this_move_value
					end
				end --if affordable
			end --if not already built
		end --for each candidate building

		--TODO- if planning to use castle, ...
		return best_move_value
	elseif bldgref.actions then
				--is this a building I want?
				local best_value = 0
				for j,action in ipairs(bldgref.actions) do
					local this_move_value = evaluate_pickup_action(r, action)
					if this_move_value > best_value then
						best_value = this_move_value
					end
				end --next action at this building
				return best_value
	end --has actions

	--do not want to place a worker here
	return -1
end

function ai.select_move()
	local whoami = current_turn()
	local me = state.players[whoami]
	local r = me.resources

	plan = build_plan(me)

	local ph = current_phase()
	if ph == PHASE_WORKER_PLACEMENT then
		return ai.select_worker_placement_move(whoami, me)
	elseif ph == PHASE_PICKUP_PROFIT then
		local bldgref = state.buildings[state.active_building].ref
		assert(bldgref.actions)
		return choose_pickup_action(r, bldgref)
	elseif ph == PHASE_CASTLE_BUILD then
		--don't want to use gold to contribute to castle
		local tg = r.gold
		r.gold = 0
		local mvs = available_moves_for_castle()
		r.gold = tg
		--any other players in castle after me?
		local more_players = #state.buildings[CASTLE_POSITION].occupants > state.turn
		local best_contribution_so_far = state.best_contribution_amount
		local function size_of_contribution(movestr)
			if movestr == "castle nobuild" then
				return 0
			else
				return math.floor((string.len(movestr) - string.len("castle ")) / 3)
			end
		end
		local function contribute_at_most(maximum)
			for i,movestr in ipairs(mvs) do
				if size_of_contribution(movestr) <= maximum then
					return movestr
				end
			end
		end
		local my_best_contribution = size_of_contribution(mvs[1])
		if not more_players and state.castle_phase <= 2 then
			--only contribute the minimum needed to earn the favor
			if my_best_contribution > best_contribution_so_far then
				return contribute_at_most(best_contribution_so_far + 1)
			else
				return contribute_at_most(1)
			end
		end
		--contribute maximum possible
		return mvs[1]
	elseif ph == PHASE_SELECT_BUILDING then
		return ai.select_builder_move(whoami, me, r)
	elseif ph == PHASE_MOVE_PROVOST then
		return ai.select_provost_move(whoami, me)
	else
		--assert(false, "current phase is "..ph)
	end
	return ai2_select()
end

function ai.select_provost_move(whoami, me)
	local movement_is_free = false

	if state.active_building == MERCHANT_GUILD_POSITION then
		movement_is_free = true
	end

	local i_am_last = false
	--check whether anyone after me on the bridge has money
	local jj = #state.order
	local bridge_occupants = state.buildings[BRIDGE_POSITION].occupants
	while bridge_occupants[jj] ~= whoami and state.players[bridge_occupants[jj]].resources.money == 0 and jj > 1 do
		jj = jj - 1
	end
	if bridge_occupants[jj] == whoami then
		i_am_last = true
	end

	--determine position of last building I am in
	if plan.last_of_my_workers > plan.last_of_opponent_workers then
		--want to move it forward as far as possible
		if movement_is_free then
			return "provost 3"
		end
		local delta = plan.last_of_my_workers - state.provost
		if delta > 0 and delta <= 3 and me.resources.money >= delta then
			return string.format("provost %d", delta)
		end
		return "noaction"
	else
		--want to move it backward
		if movement_is_free and state.provost - plan.last_of_my_workers > 4 then
			return "provost -3"
		end

		local delta = state.provost - plan.last_of_opponent_workers
		if i_am_last and delta >= 0 and delta < 3 and me.resources.money > delta + 1 then
			return string.format("provost %d", -(delta + 1))
		end
		return "noaction"
	end
end

function ai.select_worker_placement_move(whoami, me)
	if me.resources.money < 1 then
		--automatic
		return "pass"
	end
	local r = plan.resources_at_castle
	local estimated_contribution = estimate_maximum_contribution(r)
	if estimated_contribution >= 1
		and #state.buildings[CASTLE_POSITION].occupants == 0
		and me.resources.money >= get_cost_to_place_worker(whoami, state.buildings[CASTLE_POSITION])
		then
		return string.format("claim %d",CASTLE_POSITION)
	end
	if estimated_contribution >= 3
		and not plan.on_castle
		and me.resources.money >= get_cost_to_place_worker(whoami, state.buildings[CASTLE_POSITION])
		then
		return string.format("claim %d", CASTLE_POSITION)
	end
	if me.resources.money < WANT_MONEY_THRESHOLD
		and #state.buildings[TRADING_POST_POSITION].occupants == 0
		and me.resources.money >= get_cost_to_place_worker(whoami, state.buildings[TRADING_POST_POSITION])
		then
		return string.format("claim %d", TRADING_POST_POSITION)
	end
	local best_move = "pass"
	local best_move_value = 0
	for i = 1, state.bailiff + 3 do
		local bldgref = state.buildings[i].ref
		if bldgref and #state.buildings[i].occupants == 0
			and me.resources.money >= get_cost_to_place_worker(whoami, state.buildings[i])
			then
			local likelihood_of_activation = math.min(plan.expected_provost + 3 - i, 6) / 6
			local this_move_value = evaluate_worker_placement_choice(me, i)
					* likelihood_of_activation
			if this_move_value > best_move_value then
				best_move = string.format("claim %d", i)
				best_move_value = this_move_value
			end
		end --if able to place worker
	end --next building
	return best_move
end

function ai.select_builder_move(whoami, me, r)
	local bldgref = state.buildings[state.active_building].ref
	assert(bldgref)

	if bldgref == CARPENTER1 or bldgref == CARPENTER2 then
		local candidates = WOOD_BUILDINGS
		local best_move = "noaction"
		local best_move_value = 0
		for i = 1,5 do
			if not is_built(candidates[i]) then
				local cost_str = candidates[i].cost
				local cost = resource_bag.parse(cost_str)
				if r.food >= cost.food
					and r.wood >= cost.wood
					and r.stone >= cost.stone
					and r.cloth >= cost.cloth
					and r.gold >= cost.gold
					then
					--able to afford this building
					local this_move_value = estimate_value_of_building(candidates[i])
					if this_move_value > best_move_value then
						best_move_value = this_move_value
						best_move = string.format("build wood %d %s", i, cost_str)
					end
				end
			end --if not build already
		end --loop
		return best_move
	elseif bldgref == MASON then
		local candidates = STONE_BUILDINGS
		local best_move = "noaction"
		local best_move_value = 0
		for i = 1,3 do
			if not is_built(candidates[i]) then
				local cost_str = candidates[i].cost
				local cost = resource_bag.parse(cost_str)
				if r.food >= cost.food
					and r.wood >= cost.wood
					and r.stone >= cost.stone
					and r.cloth >= cost.cloth
					and r.gold >= cost.gold
					then
					--able to afford this building
					local this_move_value = estimate_value_of_building(candidates[i])
					if this_move_value > best_move_value then
						best_move_value = this_move_value
						best_move = string.format("build stone %d %s", i, cost_str)
					end
				end
			end --if not build already
		end --loop
		return best_move
	end

	return "noaction"
end

return ai
