local WANT_MONEY_THRESHOLD = 6
local ai = {}

function ai.select_move()
	local whoami = current_turn()
	local me = state.players[whoami]
	local r = me.resources

	local ph = current_phase()
	if ph == PHASE_WORKER_PLACEMENT then
		if me.resources.money < 1 then
			--automatic
			return "pass"
		end
		if r.food >= 1 and
			((r.wood >= 1 and r.stone >= 1)
				or
			 (r.wood >= 1 and r.cloth >= 1)
			 	or
			 (r.stone >= 1 and r.cloth >= 1))
			and #state.buildings[CASTLE_POSITION].occupants == 0
			and me.resources.money >= get_cost_to_place_worker(whoami, state.buildings[CASTLE_POSITION])
			then
			return string.format("claim %d",CASTLE_POSITION)
		end
		if state.bailiff >= GOLDMINE_POSITION
			and #state.buildings[GOLDMINE_POSITION].occupants == 0
			and me.resources.money >= get_cost_to_place_worker(whoami, state.buildings[GOLDMINE_POSITION])
			then
			return string.format("claim %d", GOLDMINE_POSITION)
		end
		if me.resources.money < WANT_MONEY_THRESHOLD
			and #state.buildings[TRADING_POST_POSITION].occupants == 0
			and me.resources.money >= get_cost_to_place_worker(whoami, state.buildings[TRADING_POST_POSITION])
			then
			return string.format("claim %d", TRADING_POST_POSITION)
		end
		local best_move = "pass"
		local best_move_value = 0
		for i = BRIDGE_POSITION+1, state.bailiff do
			local bldgref = state.buildings[i].ref
			if bldgref and bldgref.actions
				and #state.buildings[i].occupants == 0
				and me.resources.money >= get_cost_to_place_worker(whoami, state.buildings[i])
				then
				--is this a building I want?
				for j,action in ipairs(bldgref.actions) do
					local payout = nil
					if action == "take 2s,1c" then
						payout = { 0,0,2,1 }
					elseif action == "take 2f,1c" then
						payout = { 2,0,0,1 }
					elseif action == "take 2w,1f" then
						payout = { 1,2,0,0 }
					elseif action == "take 2f" then
						payout = { 2,0,0,0 }
					elseif action == "take 2w" then
						payout = { 0,2,0,0 }
					elseif action == "take 2s" then
						payout = { 0,0,2,0 }
					elseif action == "take 2c" then
						payout = { 0,0,0,2 }
					elseif action == "take 1f" then
						payout = { 1,0,0,0 }
					elseif action == "take 1w" then
						payout = { 0,1,0,0 }
					elseif action == "take 1s" then
						payout = { 0,0,1,0 }
					elseif action == "take 1c" then
						payout = { 0,0,0,1 }
					end
					if payout then
						local this_move_value =
							payout[1] / (r.food + 1) +
							payout[2] / (r.wood + 1) +
							payout[3] / (r.stone + 1) +
							payout[4] / (r.cloth + 1)
						if this_move_value > best_move_value then
							best_move = string.format("claim %d", i)
							best_move_value = this_move_value
						end
					end --if action is payout
				end --next action at this building
			end --if able to place worker
		end --next building
		return best_move
	elseif ph == PHASE_PICKUP_PROFIT then
		local bldgref = state.buildings[state.active_building].ref
		assert(bldgref.actions)
		if #bldgref.actions == 1 then
			return bldgref.actions[1]
		end
		--normally we prefer the first action; here are some special cases
		--where we select the second action
		if bldgref.actions[2] == "take 1c" and r.food >= 2 and (r.cloth == 0 and r.wood + r.stone <= 1) then
			return bldgref.actions[2]
		elseif bldgref.actions[2] == "take 1f" and r.food == 0 and (r.wood + r.stone + r.cloth >= 3) then
			return bldgref.actions[2]
		else
			return bldgref.actions[1]
		end
	elseif ph == PHASE_CASTLE_BUILD then
		--don't want to use gold to contribute to castle
		local tg = r.gold
		r.gold = 0
		local mvs = available_moves_for_castle()
		r.gold = tg
		--first generated move is largest contribution
		return mvs[1]
	else
		--assert(false, "current phase is "..ph)
	end
	return ai2_select()
end

return ai
