local WANT_MONEY_THRESHOLD = 6
local VALUE_OF_CARPENTER = 2
local ai = {}
local plan = {
		on_carpenter = false, --true if worker placed on a carpenter
		on_castle = false,    --true if worker placed on castle
		resources_at_castle = {}, --what resources expect to have at castle
	}

local function build_plan(me)
	local plan = {}

	local r = {
		food = me.resources.food,
		wood = me.resources.wood,
		stone = me.resources.stone,
		cloth = me.resources.cloth,
		gold = me.resources.gold,
		}

	local function i_am_here(position)
		for i,pid in ipairs(state.buildings[position].occupants) do
			if state.players[pid] == me then
				return true
			end
		end
		return false
	end

	--find the workers I've already placed
	for i = BRIDGE_POSITION+1, state.bailiff do
		if i_am_here(i) then
			local bldgref = state.buildings[i].ref
			if bldgref == CARPENTER1 or bldgref == CARPENTER2 then
				r.food = r.food - 1
				r.wood = r.wood - 1
				plan.on_carpenter = true
			end
		end
	end --next building
	if i_am_here(CASTLE_POSITION) then
		plan.on_castle = true
	end

	plan.resources_at_castle = r
	return plan
end

local function evaluate_worker_placement_choice(me, bldgref)
	local r = me.resources
	if bldgref == CARPENTER1 or bldgref == CARPENTER2 then
		if r.food >= 1 and r.wood >= 1
			and not plan.on_carpenter
			and not plan.on_castle
			then
			if not is_built(WOOD_BUILDINGS[1]) 
				or not is_built(WOOD_BUILDINGS[2])
				or not is_built(WOOD_BUILDINGS[3])
				or not is_built(WOOD_BUILDINGS[4])
				then
				return VALUE_OF_CARPENTER
			end
		end
		return -1
	elseif bldgref.actions then
				--is this a building I want?
				for j,action in ipairs(bldgref.actions) do
					local payout = nil
					if action == "take 2s,1c" then
						payout = { 0,0,2,1 }
					elseif action == "take 2f,1c" then
						payout = { 2,0,0,1 }
					elseif action == "take 2w,1f" then
						payout = { 1,2,0,0 }
					elseif action == "take 2f" then
						payout = { 2,0,0,0 }
					elseif action == "take 2w" then
						payout = { 0,2,0,0 }
					elseif action == "take 2s" then
						payout = { 0,0,2,0 }
					elseif action == "take 2c" then
						payout = { 0,0,0,2 }
					elseif action == "take 1f" then
						payout = { 1,0,0,0 }
					elseif action == "take 1w" then
						payout = { 0,1,0,0 }
					elseif action == "take 1s" then
						payout = { 0,0,1,0 }
					elseif action == "take 1c" then
						payout = { 0,0,0,1 }
					end
					if payout then
						local this_move_value =
							payout[1] / (r.food + 1) +
							payout[2] / (r.wood + 1) +
							payout[3] / (r.stone + 1) +
							payout[4] / (r.cloth + 1)
						return this_move_value
					end --if action is payout
				end --next action at this building
	end --has actions

	--do not want to place a worker here
	return -1
end

function ai.select_move()
	local whoami = current_turn()
	local me = state.players[whoami]
	local r = me.resources

	plan = build_plan(me)

	local ph = current_phase()
	if ph == PHASE_WORKER_PLACEMENT then
		if me.resources.money < 1 then
			--automatic
			return "pass"
		end
		r = plan.resources_at_castle
		if r.food >= 1 and
			((r.wood >= 1 and r.stone >= 1)
				or
			 (r.wood >= 1 and r.cloth >= 1)
			 	or
			 (r.stone >= 1 and r.cloth >= 1))
			and #state.buildings[CASTLE_POSITION].occupants == 0
			and me.resources.money >= get_cost_to_place_worker(whoami, state.buildings[CASTLE_POSITION])
			then
			return string.format("claim %d",CASTLE_POSITION)
		end
		if state.bailiff >= GOLDMINE_POSITION
			and #state.buildings[GOLDMINE_POSITION].occupants == 0
			and me.resources.money >= get_cost_to_place_worker(whoami, state.buildings[GOLDMINE_POSITION])
			then
			return string.format("claim %d", GOLDMINE_POSITION)
		end
		if me.resources.money < WANT_MONEY_THRESHOLD
			and #state.buildings[TRADING_POST_POSITION].occupants == 0
			and me.resources.money >= get_cost_to_place_worker(whoami, state.buildings[TRADING_POST_POSITION])
			then
			return string.format("claim %d", TRADING_POST_POSITION)
		end
		local best_move = "pass"
		local best_move_value = 0
		for i = BRIDGE_POSITION+1, state.bailiff do
			local bldgref = state.buildings[i].ref
			if bldgref and #state.buildings[i].occupants == 0
				and me.resources.money >= get_cost_to_place_worker(whoami, state.buildings[i])
				then
				local this_move_value = evaluate_worker_placement_choice(me, bldgref)
				if this_move_value > best_move_value then
					best_move = string.format("claim %d", i)
					best_move_value = this_move_value
				end
			end --if able to place worker
		end --next building
		return best_move
	elseif ph == PHASE_PICKUP_PROFIT then
		local bldgref = state.buildings[state.active_building].ref
		assert(bldgref.actions)
		if #bldgref.actions == 1 then
			return bldgref.actions[1]
		end
		--normally we prefer the first action; here are some special cases
		--where we select the second action
		if bldgref.actions[2] == "take 1c" and r.food >= 2 and (r.cloth == 0 and r.wood + r.stone <= 1) then
			return bldgref.actions[2]
		elseif bldgref.actions[2] == "take 1f" and r.food == 0 and (r.wood + r.stone + r.cloth >= 3) then
			return bldgref.actions[2]
		else
			return bldgref.actions[1]
		end
	elseif ph == PHASE_CASTLE_BUILD then
		--don't want to use gold to contribute to castle
		local tg = r.gold
		r.gold = 0
		local mvs = available_moves_for_castle()
		r.gold = tg
		--first generated move is largest contribution
		return mvs[1]
	elseif ph == PHASE_SELECT_BUILDING then
		if r.wood >= 1 and r.food >= 1 then
			if not is_built(WOOD_BUILDINGS[1]) then
				return string.format("build wood-2w 1w,1f na")
			elseif not is_built(WOOD_BUILDINGS[2]) then
				return string.format("build wood-2f 1w,1f na")
			elseif not is_built(WOOD_BUILDINGS[3]) then
				return string.format("build wood-2c 1w,1f na")
			elseif not is_built(WOOD_BUILDINGS[4]) then
				return string.format("build wood-2s 1w,1f na")
			end
		end
	else
		--assert(false, "current phase is "..ph)
	end
	return ai2_select()
end

return ai
