local ai = {}
ai.TUNABLE = {}
ai.TUNABLE.ALPHA = 0  --should be >= 0

--picks a move probabilistically
local function pick_move(move_scores, max_score)
	local alpha = ai.TUNABLE.ALPHA
	local adj_scores = {}
	local sum_adj = 0.0
	for m,s in pairs(move_scores) do
		adj_scores[m] = 1 / (alpha + max_score - s)
		sum_adj = sum_adj + adj_scores[m]
	end
	local rand = math.random() * sum_adj
	local result_m
	for m,v in pairs(adj_scores) do
		if rand >= 0 then
			result_m = m
		end
		rand = rand - v
	end
	return result_m
end

function ai.select_move()
	ai.context = {
		whoami = current_turn(),
		}
	if ai.VERBOSE then
		print("from current board state, we estimate:")
		local est_scores = ai.estimate_final_scores()
		for pid,s in ipairs(est_scores) do
			print(string.format("  player %s: %-20s %.1f",
				pid, PLAYERS_INFO[pid].name, s))
		end
		print("we must pick from the following moves:")
	end

	local saved_verbose = ai.VERBOSE
	ai.VERBOSE = false
	local saved_state = state
	local move_scores = {}
	local best
	local best_score = -math.huge
	for i,m in ipairs(all_available_moves()) do
		state = clone_game_state(saved_state)
		do_move(m)
		local score = ai.evaluate_position()
		move_scores[m] = score
		if score > best_score then
			best = m
			best_score = score
		end
		if saved_verbose then
			print(string.format("move %s is worth %.1f", m, score))
		end
	end
	state = saved_state
	ai.VERBOSE = saved_verbose
	if ai.TUNABLE.ALPHA == 0 then
		return best
	else
		return pick_move(move_scores, best_score)
	end
end

local function evaluate_scores(est_scores, whoami)
	local sum = 0.0
	for i,score in ipairs(est_scores) do
		sum = sum + score
	end
	local avg_opponent_score = (sum - est_scores[whoami])
			/ (#est_scores - 1)
	return est_scores[whoami] - avg_opponent_score
end

function ai.evaluate_position(whoami)
	whoami = whoami or ai.context.whoami

	local est_scores = ai.estimate_final_scores()
	return evaluate_scores(est_scores, whoami)
end

load_ai_mixin(ai, "roundsleft_helper.inc")
load_ai_mixin(ai, "valuation_jl1.inc")

return ai
