local ai = ai

assert(ai)

--In this "mix-in", we try to correctly quantify the value of having
--unspent resources. This includes money, food, wood, stone, cloth,
--and gold. The number returned is, theoretically, the number of victory
--points these unspent resources will translate to.

--@param a value of first item
--@param b marginal value of second item
--@param c marginal value of nth item (for a large n)
--@return a function that will calculate the value
local function decaying_rewards(a,b,c)
	local y1 = a
	local y2 = y1 + b
	
	local bb = ((2 * y2 - 4 * c) - (2 * y1 - 2 * c))
		/ ((2 * y1 - 2 * c) - (y2 - 2 * c))
	local aa = (y1 - c) * (bb + 1)
	return
	function(n)
		return c * n + aa * n / (bb + n)
	end
end

local value_of_money = decaying_rewards(2.0, 1.5, .25)
local value_of_gold = decaying_rewards(9.0, 7.5, 3.0)
local value_of_stone = decaying_rewards(3.5, 3.0, .25)

--@param r a resources struct
--
function ai.value_of_resources(r)
	--'a' is in [0,1], represents how far along in the game we are
	--we use this to prefer wood near the beginning of the game, but
	--stone towards the end of the game.
	local a = math.min(1, (state.round - 1) / 12)

	return
		decaying_rewards(2.5-a, .5, .33)(r.food) +
		decaying_rewards(3.5-a, .5, .33)(r.wood) +
		value_of_stone(r.stone) +
		decaying_rewards(3.4-a, .5, .33)(r.cloth) +
		value_of_gold(r.gold) +
		value_of_money(r.money)
end


return ai
