local ai = ai

assert(ai)

local TYPICAL_NUMBER_OF_ROUNDS = 14.5

--if a player has this many cubes or more, we assume the player will
--try to soon grab castle blocks
local N_CUBES_THRESHOLD = 6

--how many spaces per round we expect the bailiff to move
local BAILIFF_SPEED = 1.615

--use these tunables to calculate how fast the castle is filling up
local CASTLE_RATE_X0 = 2.493562
local CASTLE_RATE_X1 = 0.171395

function ai.estimate_rounds_left()
	--copied from nn.lua, 2010-09-11
	if state.castle_phase > 3 then
		return 0
	end

	local bailiff_left = CASTLE_BAILIFF_THRESHOLDS[3] - state.bailiff 
	local castle_left = CASTLE_CONTRIBUTION_THRESHOLDS[3] - state.contribution_count

	if bailiff_left <= 1 then
		return 0
	end

	--bailiff movement rate:
	--assumed constant through game
	--should probably be dependent on whether there are buildings
	--out beyond where the bailiff is
	local bailiff_speed = BAILIFF_SPEED
	local est_bailiff_rounds_left = bailiff_left / bailiff_speed

	--ready for delivery:
	local n_cubes_in_stock = 0
	for pid,player in pairs(state.players) do
		local r = player.resources
		local n_cubes = r.food + r.wood + r.stone + r.cloth
		n_cubes_in_stock = n_cubes_in_stock
			+ math.max(0, n_cubes - N_CUBES_THRESHOLD)
	end
	castle_left = math.max(1, castle_left - n_cubes_in_stock / 3)

	--castle block rates:
	--in first five rounds, estimate 1.2 blocks per round
	--in rounds 6-10, estimate 2 blocks per round
	--in rounds 11-15, estimate 2.8 blocks per round
	--TODO- adjust this based on what buildings are available
	local castle_block_speed = CASTLE_RATE_X1*state.round + CASTLE_RATE_X0
	local est_castle_rounds_left = castle_left / castle_block_speed

	local est_rounds_left = math.min(
			est_castle_rounds_left,
			est_bailiff_rounds_left)

	local est1 = TYPICAL_NUMBER_OF_ROUNDS - (state.round - 1)
	local est2 = est_rounds_left

	local portion = 1 - math.max(0,est_rounds_left - 1) / TYPICAL_NUMBER_OF_ROUNDS

	return est1 * (1-portion) + est2 * (portion)
end

function ai.learn_rounds_left(real_rounds_left, which_ended)

	--copied from nn.lua, 2010-09-11
	if state.castle_phase > 3 then
		return {}
	end

	which_ended = which_ended or "both"

	local bailiff_left = CASTLE_BAILIFF_THRESHOLDS[3] - state.bailiff 
	local castle_left = CASTLE_CONTRIBUTION_THRESHOLDS[3] - state.contribution_count

	local bailiff_portion = (state.bailiff - BAILIFF_START_POSITION)
		/ (CASTLE_BAILIFF_THRESHOLDS[3] - BAILIFF_START_POSITION)
	local castle_portion = state.contribution_count / CASTLE_CONTRIBUTION_THRESHOLDS[3]
	local avg_portion = (bailiff_portion + castle_portion) / 2.0

	--bailiff movement rate:
	--assumed constant through game
	--should probably be dependent on whether there are buildings
	--out beyond where the bailiff is
	local bailiff_speed = BAILIFF_SPEED
	local est_bailiff_rounds_left = bailiff_left / bailiff_speed

	--ready for delivery:
	local n_cubes_in_stock = 0
	for pid,player in pairs(state.players) do
		local r = player.resources
		local n_cubes = r.food + r.wood + r.stone + r.cloth
		n_cubes_in_stock = n_cubes_in_stock
			+ math.max(0, n_cubes - N_CUBES_THRESHOLD)
	end
	castle_left = math.max(1, castle_left - n_cubes_in_stock / 3)

	--castle block rates:
	--in first five rounds, estimate 1.2 blocks per round
	--in rounds 6-10, estimate 2 blocks per round
	--in rounds 11-15, estimate 2.8 blocks per round
	--TODO- adjust this based on what buildings are available
	local castle_block_speed = CASTLE_RATE_X1 * state.round + CASTLE_RATE_X0
	local est_castle_rounds_left = castle_left / castle_block_speed

	local w_ecrl = 1
	local w_ebrl = 1
	if which_ended == "bailiff" then
		w_ecrl = 0
	elseif which_ended == "castle" then
		w_ebrl = 0
	end

	local est_rounds_left =
			w_ecrl * est_castle_rounds_left +
			w_ebrl * est_bailiff_rounds_left

	local est1 = TYPICAL_NUMBER_OF_ROUNDS - (state.round - 1)
	local est2 = est_rounds_left

--avg_portion=1
	local my_est = est1 * (1-avg_portion) + est2 * (avg_portion)
	local err = real_rounds_left - my_est

	local adj_tnor = err * (1-avg_portion)
	local adj_erl = err * avg_portion
	local adj_ecrl = adj_erl * (w_ecrl / (w_ecrl + w_ebrl))
	local adj_ebrl = adj_erl * (w_ebrl / (w_ecrl + w_ebrl))
	local adj_bs = adj_ebrl * bailiff_left * -1 / math.pow(bailiff_speed,2)
	local adj_cs = adj_ecrl * castle_left * -1 / math.pow(castle_block_speed,2)

	local adj_crx0 = adj_cs * 1
	local adj_crx1 = adj_cs * state.round

	return {
		adj_rounds_left = err,
		adj_est_castle_rounds_left = adj_ecrl,
		adj_castle_speed = adj_cs,
		TYPICAL_NUMBER_OF_ROUNDS = adj_tnor,
		BAILIFF_SPEED = adj_bs,
		CASTLE_RATE_X0 = adj_crx0,
		CASTLE_RATE_X1 = adj_crx1,
		}, my_est
end

return ai
