local ai = ai
assert(ai)
assert(ai.TUNABLE)
ai.TUNABLE.VALUE_OF_FAVOR = 3.6
ai.TUNABLE.VALUE_OF_CONTRIBUTION = 2.0
ai.TUNABLE.VALUE_OF_FUTURE_CONTRIBUTION = 1.43
ai.TUNABLE.VALUE_OF_FIRST_BLOCK = 3
ai.TUNABLE.VALUE_OF_ORDER = -.3
ai.TUNABLE.VALUE_OF_INCOME_PER_ROUND = 1.3
ai.TUNABLE.VALUE_PER_ROUND = 1.0
ai.TUNABLE.VALUE_OFFSET = 0
ai.TUNABLE.VALUE_OF_ARCHITECT = 10
ai.TUNABLE.VALUE_OF_VP_FAVORS = 0.77
ai.TUNABLE.VALUE_OF_MONEY_FAVORS = 0.55
ai.TUNABLE.VALUE_OF_RESOURCE_FAVORS = 0.81
ai.TUNABLE.VALUE_OF_BUILD_FAVORS = 1.2
ai.TUNABLE.VALUE_OF_VP_FAVORS_PER_ROUND = 0
ai.TUNABLE.VALUE_OF_MONEY_FAVORS_PER_ROUND = 0
ai.TUNABLE.VALUE_OF_RESOURCE_FAVORS_PER_ROUND = 0
ai.TUNABLE.VALUE_OF_BUILD_FAVORS_PER_ROUND = 0
ai.TUNABLE.VALUE_OF_INN = 0
ai.TUNABLE.VALUE_FIRST_RESIDENCE = 1.0
ai.TUNABLE.VALUE_PER_RESIDENCE = 0.1
ai.TUNABLE.VALUE_OF_RESOURCE = {
	money = { 8.0, .40,  21, 6.8,-.33,  22 },
	food  = { 8.3, .40,  22, 8.0,   0,  24 },
	wood  = { 8.1, .38,  22, 9.4, 1.3,  23 },
	stone = { 8.4, .39,  22, 7.2, .65,  23 },
	cloth = { 8.1, .38,  22, 9.2, 1.3,  23 },
	gold  = { 8.5, .40, 200, 8.0, .79, 200 },
	}
ai.TUNABLE.VALUE_OF_BUILDING = {
	}
ai.TUNABLE.VALUE_OF_BUILDING_POTENTIAL = {
	["wood-2w"] = 2,
	["wood-2f"] = 2,
	["wood-2c"] = 2,
	["wood-2s"] = 2,
	mason = 4,
	["wood-peddler"] = 4,
	lawyer = 4,
	["wood-market"] = 4,
	["stone-ffc"] = 4,
	["stone-ssc"] = 4,
	["stone-wwf"] = 4,
	alchemist = 6,
	architect = 6,
	bank = 6,
	church = 6,
	jeweler = 6,
	tailor = 6,
	statue = 10,
	weaver = 12,
	library = 10,
	hotel = 16,
	granary = 10,
	theater = 17,
	college = 17,
	monument = 20,
	cathedral = 25,
	}
ai.TUNABLE.VALUE_OF_COMBO = {
	gold_stone_stone = 5
	}

local ALL_BUILDING_IDS = { }
for i,bldgref in ipairs(WOOD_BUILDINGS) do
	table.insert(ALL_BUILDING_IDS, bldgref.id)
end
for i,bldgref in ipairs(STONE_BUILDINGS) do
	table.insert(ALL_BUILDING_IDS, bldgref.id)
end
for i,bldgref in ipairs(PRESTIGE_BUILDINGS) do
	table.insert(ALL_BUILDING_IDS, bldgref.id)
end
for i,bldg_id in ipairs(ALL_BUILDING_IDS) do
	ai.TUNABLE.VALUE_OF_BUILDING[bldg_id]
		= (ai.TUNABLE.VALUE_OF_BUILDING[bldg_id] or 0)
	ai.TUNABLE.VALUE_OF_BUILDING_POTENTIAL[bldg_id]
		= (ai.TUNABLE.VALUE_OF_BUILDING_POTENTIAL[bldg_id] or 0)
end


--- Guess how many castle blocks this player can build.
-- @param r a "resources" struct
-- @return a whole number
function ai.estimate_maximum_contribution(r)

	--patterns like '3f,2w,2s,2c' (3 contributions) are difficult
	--for the simple "max" function found below, so we first
	--calculate how many batches of '3f,2w,2s,2c' there are,
	--then use the simple "max" function for the remainder
	--thanks to zhouyan.cn for this algorithm...

	local a0 = r.food
	local a1 = r.wood
	local a2 = r.stone
	local a3 = r.cloth

	local b = math.min(
		math.floor(a0 / 3),
		math.floor(a1 / 2),
		math.floor(a2 / 2),
		math.floor(a3 / 2)
		)
	a0 = a0 - b * 3
	a1 = a1 - b * 2
	a2 = a2 - b * 2
	a3 = a3 - b * 2

	return b * 3 + math.max(
		math.min(a0, a1, a2 + a3),
		math.min(a0, a2, a1 + a3),
		math.min(a0, a3, a1 + a2)
		)
end

local function try_to_find(r, rsrc)
	local a = r._alternate
	local b = r
	while a do
		for i,bag in ipairs(a.data) do
			if bag[rsrc] > 0 then
				--found it
				r.food = r.food + bag.food
				r.wood = r.wood + bag.wood
				r.stone = r.stone + bag.stone
				r.cloth = r.cloth + bag.cloth
				r.money = r.money + bag.money
				if b == r then
					b._alternate = a.next
				else
					b.next = a.next
				end
				return true
			end
		end
		b = a
		a = a.next
	end
	--not found
	return false
end

--A note on "alternates".
--When workers get activated, sometimes there are choices.
--One choice becomes the main choice, and the player_r structure is
--updated to reflect that choice.
--Other choices are stored as "alternates", a linked list off of the
--player_r[pid] structure.
--
--So, e.g. say the player with 2f,1w has a worker on a farm
--giving 1f/1c. 1f is selected as the "main choice", so the
--player_r struct is:
--   +--------+
--   | 3f,1w  |
--   +--------+
--             \-----alternate---> +--------+
--                                 | -1f,1c |
--                                 +--------+
--
--How about that same player now has a worker on farm and forest.
--Now there are two places where alternate choices could be made,
--i.e. there are four possibilities:
--          2f,1w + 1f + 1f  ==  4f,1w
--          2f,1w + 1f + 1w  ==  3f,2w
--          2f,1w + 1c + 1f  ==  3f,1w,1c
--          2f,1w + 1c + 1w  ==  2f,2w,1c
--          
--   +-------+                  +--------+
--   | 4f,1w |----alternate---> | -1f,1w |
--   +-------+                  +--------+
--                                   V next alternate
--                              +--------+
--                              | -1f,1c |
--                              +--------+
--                                   V next alternate
--                              +-----------+
--                              | -2f,1w,1c |
--                              +-----------+
--
--There are times when some alternative is only available if a
--certain earlier choice is made. E.g. if you take cloth at the 1f/1c
--farm, then you can build lawyer at the carpenter.
--
--E.g. player has 2f,1w; worker on 1f/1c farm; worker on carpenter.
--Default choice is 1f at farm, and build mason at carpenter.
--
-- +-----------+              +--------+              +----------------+
-- | 2f,+mason |--alternate-> | -1f,1c |--alternate-> | -mason,+lawyer |
-- +-----------+              +--------+              +----------------+
--                                V next
--                            +----------------+
--                            | -mason,+quarry |
--                            +----------------+
--                                V next
--                            +-----------------------+
--                            | -1f,1c,-mason,+quarry |
--                            +-----------------------+
--
--I.e. can build quarry instead of mason irregardless of whether we get
--cloth at the farm. But to build a lawyer, we must get cloth not food
--at the farm.
--
local function adjust_obtained(r, bldgref)
	if bldgref.actions then
		if #bldgref.actions == 1 then
			local movestr = bldgref.actions[1]
			local rsrc_str = string.match(movestr, "^take (.*)$")
			if rsrc_str then
				local bag = resource_bag.parse(rsrc_str)
				r.food = r.food + bag.food
				r.wood = r.wood + bag.wood
				r.stone = r.stone + bag.stone
				r.cloth = r.cloth + bag.cloth
				r.gold = r.gold + bag.gold
				r.money = r.money + bag.money
			end
		else
			local alternates = {}
			for i=1,#bldgref.actions do
				local str2 = string.match(bldgref.actions[i], "^take (.*)$")
				if str2 then
					local bag2 = resource_bag.parse(str2)
					table.insert(alternates, bag2)
				end
			end
			r._alternate = {
				next = r._alternate,
				data = alternates,
				}
			return
		end
	end

	if bldgref.id == "fixed-peddler" then
		if r.money >= 2 then
			r.money = r.money - 2
			local alternates = {}
			table.insert(alternates, resource_bag.parse("1f"))
			table.insert(alternates, resource_bag.parse("1w"))
			table.insert(alternates, resource_bag.parse("1s"))
			table.insert(alternates, resource_bag.parse("1c"))
			r._alternate = {
				next = r._alternate,
				data = alternates,
				}
		end
	elseif bldgref.id == "wood-peddler" then
		if r.money == 0 then
			try_to_find(r, "money")
		end
		if r.money >= 4 then
			--assume player will spend two coins
			r.money = r.money - 2
			local alternates = {}
			table.insert(alternates, resource_bag.parse("1f"))
			table.insert(alternates, resource_bag.parse("1w"))
			table.insert(alternates, resource_bag.parse("1s"))
			table.insert(alternates, resource_bag.parse("1c"))
			r._alternate = {
				next = r._alternate,
				data = alternates,
				}
			r._alternate = {
				next = r._alternate,
				data = alternates,
				}
		elseif r.money >= 1 then
			--assume player will spend only one coin
			r.money = r.money - 1
			local alternates = {}
			table.insert(alternates, resource_bag.parse("1f"))
			table.insert(alternates, resource_bag.parse("1w"))
			table.insert(alternates, resource_bag.parse("1s"))
			table.insert(alternates, resource_bag.parse("1c"))
			r._alternate = {
				next = r._alternate,
				data = alternates,
				}
		end

	elseif bldgref.id == "wood-market" or bldgref.id == "neutral-market"
	then
		local found = {}
		if r.food >= 1 then
			r.food = r.food - 1
			table.insert(found, "1f")
		end
		if r.wood >= 1 then
			r.wood = r.wood - 1
			table.insert(found, "1w")
		end
		if r.stone >= 1 then
			r.stone = r.stone - 1
			table.insert(found, "1s")
		end
		if r.cloth >= 1 then
			r.cloth = r.cloth - 1
			table.insert(found, "1c")
		end
		if #found >= 1 then
			r.money = r.money + (bldgref.id == "wood-market" and 6 or 4)
		end
		if #found >= 2 then
			local alternates = {}
			for i in ipairs(found) do
				--each "alternate" represents the choice of which
				--three of the four resource types that do NOT get
				--sold at market
				local f1 = {}
				for j,s in ipairs(found) do
					if i ~= j then
						table.insert(f1, s)
					end
				end
				table.insert(alternates, resource_bag.parse(
					table.concat(f1, ",")))
			end
			r._alternate = {
				next = r._alternate,
				data = alternates,
				}
		end --if some choice available

	elseif bldgref.id == "lawyer" then
		if r.money == 0 then
			try_to_find(r, "money")
		end
		if r.cloth == 0 then
			try_to_find(r, "cloth")
		end
		if r.money >= 1 and r.cloth >= 1 then
			r.money = r.money - 1
			r.cloth = r.cloth - 1
			r.residences = (r.residences or 0) + 1
			r.income = r.income + 1
			r.score = r.score + 2
		end

	elseif bldgref.id == "church" then

		if r.money >= 4 then
			r.money = r.money - 4
			r.score = r.score + 5
		elseif r.money >= 2 then
			r.money = r.money - 2
			r.score = r.score + 3
		end

	elseif bldgref.id == "bank" then

		if r.money >= 5 then
			r.money = r.money - 5
			r.gold = r.gold + 2
		elseif r.money >= 2 then
			r.money = r.money - 2
			r.gold = r.gold + 1
		end

	elseif bldgref.id == "jeweler" then

		if r.gold >= 2 then
			r.gold = r.gold - 2
			r.score = r.score + 9
		elseif r.gold >= 1 then
			r.gold = r.gold - 1
			r.score = r.score + 5
		end

	elseif bldgref.id == "joust" then

		if r.cloth >= 1 and r.money >= 1 then
			r.cloth = r.cloth - 1
			r.money = r.money - 1
			r.favors = r.favors + 1
		end

	elseif bldgref.id == "tailor" then

		try_to_find(r, "cloth")
		if r.cloth >= 3 then
			r.cloth = r.cloth - 3
			r.score = r.score + 6
		elseif r.cloth >= 2 then
			r.cloth = r.cloth - 2
			r.score = r.score + 4
		end

	end

	return
end

local function consider_potential_building(r, bldgref)
	if bldgref.actions then
		local movestr = bldgref.actions[1]
		local rsrc_str = string.match(movestr, "^take (.*)$")
		if rsrc_str then
			local bag = resource_bag.parse(rsrc_str)
			r._potential = bag
		end
	end
end

local function new_order()
	local order = {}
	local found = {}
	for i,pid in ipairs(state.buildings[STABLES_POSITION].occupants) do
		table.insert(order, pid)
		found[pid] = true
	end
	for i,pid in ipairs(state.order) do
		if not found[pid] then
			table.insert(order, pid)
		end
	end
	assert(#order == #state.order)
	return order
end

local function process_castle_and_round_end(player_r, rounds_left)
	local bldg = state.buildings[CASTLE_POSITION]
	local slots_left = CASTLE_CONTRIBUTION_THRESHOLDS[3] - state.contribution_count
	local best_contrib = 0
	local best_contrib_pid
	for i,pid in ipairs(bldg.occupants) do
		if player_r[pid] then
			local r = player_r[pid]
			if r._potential then
				--TODO: consider additional resources that
				--could be collected by placing more workers
				--before castle is activated
			end
			if r.food == 0 then
				try_to_find(r, "food")
			end
			if r.wood == 0 then
				try_to_find(r, "wood")
			end
			if r.stone == 0 then
				try_to_find(r, "stone")
			end
			if r.cloth == 0 then
				try_to_find(r, "cloth")
			end
			local function deduct_contribution()
				if r.food < 1 then
					return false
				elseif r.wood >= 1 and r.stone >= 1
					and math.min(r.wood,r.stone)>=r.cloth
				then
					r.food = r.food - 1
					r.wood = r.wood - 1
					r.stone = r.stone - 1
					return true
				elseif r.wood >= 1 and r.cloth >= 1
					and math.min(r.wood,r.cloth)>=r.stone
				then
					r.food = r.food - 1
					r.wood = r.wood - 1
					r.cloth = r.cloth - 1
					return true
				elseif r.stone >= 1 and r.cloth >= 1
				then
					r.food = r.food - 1
					r.stone = r.stone - 1
					r.cloth = r.cloth - 1
					return true
				else
					return false
				end
			end
			local num_contrib = 0
			while slots_left >= 1 and deduct_contribution() do
				num_contrib = num_contrib + 1
				slots_left = slots_left - 1
			end
			if num_contrib >= 1 then
				r.contrib_now = num_contrib
				r.contrib_later = 0
				r.contrib_this_round = 1
				if num_contrib > best_contrib then
					best_contrib = num_contrib
					best_contrib_pid = pid
				end
			else
				r.score = math.max(0, r.score - 2)
			end
		end
	end
	if best_contrib_pid then
		local r = player_r[best_contrib_pid]
		r.favors = r.favors + 1
	end

	if rounds_left < 1 then
		slots_left = 0
	end
	for i,pid in ipairs(new_order()) do
		local r = player_r[pid]
		if not r.contrib_now then
			r.contrib_now = 0
			r.contrib_later = math.min(slots_left,
				ai.estimate_maximum_contribution(r))
			slots_left = slots_left - r.contrib_later
		end
	end
	if rounds_left < 3 then
		for pid,r in pairs(player_r) do
			r.contrib_later = (r.contrib_later or 0) + slots_left / #player_r
		end
	end
end

local function logisticfn(t)
	return 1 / (1 + math.exp(-t))
end
local function logistic_1der(t)
	return t * (1-t)
end

if not ai.estimate_rounds_left then
	ai.estimate_rounds_left = function()
		return math.max(0, 13-state.round)
	end
end

local function can_pay(r, cost)
	if cost == "1w,1_" then
		return (r.wood >= 1 and r.food + r.wood + r.stone + r.cloth >= 2)
	elseif cost == "1w,1c" then
		return (r.wood >= 1 and r.cloth >= 1)
	elseif cost == "1w,1f" then
		return (r.wood >= 1 and r.food >= 1)
	elseif cost == "1s,1f" then
		return (r.stone >= 1 and r.food >= 1)
	elseif cost == "1s,1c" then
		return (r.stone >= 1 and r.cloth >= 1)
	elseif cost == "1s,1w" then
		return (r.stone >= 1 and r.wood >= 1)
	else
		local c = resource_bag.parse(cost)
		return (
			r.food >= c.food and
			r.wood >= c.wood and
			r.stone >= c.stone and
			r.cloth >= c.cloth and
			r.gold >= c.gold and
			r.money >= c.money
			)
	end
end

--helper function for process_future_architects
--
-- @param can_pay_now true if the player has the resources already to
--        build this building; the player has not yet placed on the
--        carpenter/mason/architect
-- @param rounds_left the expected number of rounds remaining
--
-- @return a value between 0 and 1 indicating the ease with which a player
--         can build a building; 1 being most easy
--
local function ease_of_building(can_pay_now, rounds_left)
	if can_pay_now then
		--player already has the resources; should be easy to
		--build this building, even with no rounds left

		local MIN_EASE = .9
		return rounds_left < 1 and MIN_EASE
			or (MIN_EASE+rounds_left-1)/rounds_left
	else
		--player is still lacking resources; ease depends on
		--how many rounds still to play

		local MIN_EASE = .2
		local EASE2 = .4
		return rounds_left < 2 and MIN_EASE
			or (EASE2*2+rounds_left-2)/rounds_left
	end
end

--helper for process_future_{carpenters,masons,architects}
--
--This subroutine adds value to the "bldg_$id", "can_build_$id"
--resources on the resource structure.
--
local function future_builder(r, bldgid, can_pay_now, rounds_left,
			expected_activations_this_player,
			num_buildable)

	local k1 = "bldg_"..bldgid      --expected num. rounds owning this
	local k2 = "can_build_"..bldgid --probability of building it

	local ease = ease_of_building(can_pay_now, rounds_left)
	local prob_build = num_buildable == 0 and 0
		or ease * expected_activations_this_player / num_buildable
	r[k2] = (r[k2] or 0) + prob_build

	--consider the benefit of owning this building
	--  need at least three rounds left
	--  the current round is already too late for building.
	--  the next round is the earliest the building could be
	--  built. the round after is the earliest we can
	--  benefit from having this building
	if rounds_left >= 3 then
		local prob_per_round = prob_build / (rounds_left - 1)
		local e_r_own = prob_per_round * (rounds_left - 2) * (rounds_left - 1) / 2
		r[k1] = (r[k1] or 0) + e_r_own
	end
end

--figures out which wooden buildings are still buildable,
--and likelihood/difficulty of building each of those buildings
--
local function process_future_carpenters(pr, rounds_left)

	local buildable = {}
	for i,bldgref in ipairs(WOOD_BUILDINGS) do
		if not is_built(bldgref) then
			table.insert(buildable, {
				id = bldgref.id,
				cost = bldgref.cost,
				})	
		end
	end

	local expected_carpenter_activations = math.min(rounds_left, #buildable)
	local expected_activations_per_player = expected_carpenter_activations / #state.players

	for i,b in ipairs(buildable) do

		for pid in ipairs(state.players) do
			local r = pr[pid]

			--see if we can scrounge up wood/food to pay for the
			--building.
			--FIXME: not all wood buildings need food
			if r.wood == 0 then
				try_to_find(r, "wood")
			end
			if r.food == 0 then
				try_to_find(r, "food")
			end
			local can_pay_now = can_pay(r, b.cost)

			future_builder(r, b.id, can_pay_now, rounds_left,
				expected_activations_per_player,
				#buildable)
		end
	end
end

--figures out which stone buildings are still buildable,
--and likelihood/difficulty of building each of those buildings
--
local function process_future_masons(pr, rounds_left)

	local buildable = {}
	for i,bldgref in ipairs(STONE_BUILDINGS) do
		if not is_built(bldgref) then
			table.insert(buildable, {
				id = bldgref.id,
				cost = bldgref.cost,
				})	
		end
	end

	local expected_activations = math.min(
		(is_built(MASON) and rounds_left or math.max(0,rounds_left-2)),
		#buildable)
	local expected_activations_per_player = expected_activations / #state.players

	--TODO: if there are players advanced on the building favor track,
	--then we can expect more stone buildings to be built

	for i,b in ipairs(buildable) do

		for pid in ipairs(state.players) do
			local r = pr[pid]

			--see if we can scrounge up stone to pay for the
			--building.
			if r.stone == 0 then
				try_to_find(r, "stone")
			end
			local can_pay_now = can_pay(r, b.cost)

			future_builder(r, b.id, can_pay_now, rounds_left,
				expected_activations_per_player,
				#buildable)
		end
	end
end

--figures out which prestige buildings are still buildable,
--and likelihood/difficulty of building each of those buildings
--
local function process_future_architects(pr, rounds_left)

	--determine which prestige buildings are still available
	local buildable = {}
	for i,bldgref in ipairs(PRESTIGE_BUILDINGS) do
		if not is_built(bldgref) then
			table.insert(buildable, {
				id = bldgref.id,
				cost = bldgref.cost,
				})	
		end
	end

	--check whether certain buildings exist, and how many
	--residences each player has built
	local is_lawyer_built = false
	local is_architect_built = false
	local num_residences = {}
	for i,bldg in ipairs(state.buildings) do
		if bldg.ref == LAWYER then
			is_lawyer_built = true
		elseif bldg.ref == ARCHITECT then
			is_architect_built = true
		elseif bldg.ref == RESIDENCE then
			num_residences[bldg.owner] = (num_residences[bldg.owner] or 0) + 1
		end
	end

	--
	--estimate how many times each player is likely to build a
	--prestige building
	--

	local expected_activations = {}
	local sum_expected_activations = 0
	for pid in pairs(state.players) do
		expected_activations[pid] =
			(num_residences[pid] or 0) > 0 and math.min(rounds_left, num_residences[pid])
			or rounds_left > 4 and 1
			or 0
		if state.players[pid].build_favor >= 2 then
			expected_activations[pid] = expected_activations[pid] + 1
		elseif not is_architect_built then
			expected_activations[pid] = expected_activations[pid] / 2
		end
		sum_expected_activations = sum_expected_activations
			+ expected_activations[pid]
	end
	local adjusted_expected_activations =
		math.min(sum_expected_activations, rounds_left, #buildable)
	local adj = sum_expected_activations == 0 and 0
		or adjusted_expected_activations / sum_expected_activations
	for pid in pairs(state.players) do
		expected_activations[pid] = expected_activations[pid] * adj
	end

	for i,b in ipairs(buildable) do

		for pid in ipairs(state.players) do
			local r = pr[pid]
			local can_pay_now = can_pay(r, b.cost)

			future_builder(r, b.id,
				can_pay_now, rounds_left,
				expected_activations[pid],
				#buildable)
		end
	end
end

local function process_occupied_carpenter(occupant, player_r, rounds_left)
	
	local r = player_r[occupant]

	local buildable = {}
	for i,bldgref in ipairs(WOOD_BUILDINGS) do
		if not is_built(bldgref) and can_pay(r, bldgref.cost) then
			table.insert(buildable, {
				id = bldgref.id,
				cost = bldgref.cost,
				})	
		end
	end

	for i,b in ipairs(buildable) do
		local k1 = "bldg_"..b.id      --expected num. rounds owning this
		local k2 = "can_build_"..b.id --probability of building it

		local prob_build = #buildable == 0 and 0 or 1 / #buildable
		r[k2] = (r[k2] or 0) + prob_build

		--expected number of rounds to own this building
		local e_r_own = prob_build * math.max(0, rounds_left - 1)
		r[k1] = (r[k1] or 0) + e_r_own
	end

	r.wood = math.max(0, r.wood-1)
	r.food = math.max(0, r.food-1)
end

local function process_occupied_mason(occupant, player_r, rounds_left)
	
	local r = player_r[occupant]

	local buildable = {}
	for i,bldgref in ipairs(STONE_BUILDINGS) do
		if not is_built(bldgref) and can_pay(r, bldgref.cost) then
			table.insert(buildable, {
				id = bldgref.id,
				cost = bldgref.cost,
				})	
		end
	end

	for i,b in ipairs(buildable) do
		local k1 = "bldg_"..b.id      --expected num. rounds owning this
		local k2 = "can_build_"..b.id --probability of building it

		local prob_build = #buildable == 0 and 0 or 1 / #buildable
		r[k2] = (r[k2] or 0) + prob_build

		--expected number of rounds to own this building
		local e_r_own = prob_build * math.max(0, rounds_left - 1)
		r[k1] = (r[k1] or 0) + e_r_own
	end

	r.stone = math.max(0, r.stone-1)
	r.food = math.max(0, r.food-1)
end

local function process_occupied_architect(occupant, player_r, rounds_left)

	local r = player_r[occupant]
	if (r.residences or 0) == 0 then
		return
	end

	local buildable = {}
	for i,bldgref in ipairs(PRESTIGE_BUILDINGS) do
		if not is_built(bldgref) and can_pay(r, bldgref.cost) then
			table.insert(buildable, {
				id = bldgref.id,
				cost = bldgref.cost,
				})
		end
	end

	for i,b in ipairs(buildable) do
		local k1 = "bldg_"..b.id      --expected num. rounds owning this
		local k2 = "can_build_"..b.id --probability of building it

		local prob_build = #buildable == 0 and 0 or 1 / #buildable
		r[k2] = (r[k2] or 0) + prob_build

		--expected number of rounds to own this building
		local e_r_own = prob_build * math.max(0, rounds_left - 1)
		r[k1] = (r[k1] or 0) + e_r_own
	end

	r.gold = math.max(0, r.gold - 1)
	r.residences = r.residences - 1
	r.income = r.income - 1
	r.can_build_prestige = (r.can_build_prestige or 0) + 1

end

local function process_building(bldg, player_r, rounds_left)

	local bldgref = bldg.ref
	if bldg.occupants[1] and player_r[bldg.occupants[1]] then

		if bldgref.id == "neutral-carpenter" or bldgref.id == "fixed-carpenter"
		then
			process_occupied_carpenter(bldg.occupants[1], player_r, rounds_left)
		elseif bldgref.id == "mason" then
			process_occupied_mason(bldg.occupants[1], player_r, rounds_left)
		elseif bldgref.id == "architect" then
			process_occupied_architect(bldg.occupants[1], player_r, rounds_left)
		elseif bldgref.id == "inn" then
			if bldg.owner and player_r[bldg.owner] then
				player_r[bldg.owner].in_inn = nil
			end
			player_r[bldg.occupants[1]].in_inn = 1
		else

			adjust_obtained(player_r[bldg.occupants[1]], bldgref)
		end

	elseif bldgref then		 --unoccupied building

		if bldgref.id == "inn" then
			if bldg.owner and player_r[bldg.owner] then
				--assume current occupant of inn will stay
				player_r[bldg.owner].in_inn = 1
			end

		elseif bldgref.actions then
			for pid in ipairs(state.players) do
				consider_potential_building(player_r[pid], bldgref)
			end
		end
	end
end

function ai.estimate_final_scores(scratch_area)

	local rounds_left = ai.estimate_rounds_left()
	local portion_of_game_left = math.min(1, rounds_left / 13)

	local player_r_by_position = {}
	local player_r = {}
	for pid,player in pairs(state.players) do
		player_r[pid] = resource_bag.clone(player.resources)
		player_r[pid].score = player.score
		player_r[pid].income = player.income
		player_r[pid].vp_favor = player.vp_favor
		player_r[pid].money_favor = player.money_favor
		player_r[pid].resource_favor = player.resource_favor
		player_r[pid].build_favor = player.build_favor
		player_r[pid].favors = 0
	end
	if state.favors_active then
		local r = player_r[current_turn()]
		r.favors = r.favors + 1
	end
	for i,pid in ipairs(new_order()) do
		--turn order position much less important on last round
		--FIXME: if two or more players are advanced on building track,
		--then turn order could be important
		player_r[pid].turn_order_position = rounds_left > 0 and i or 0
	end
	local num_architects = 0

	for i = 1, CASTLE_POSITION - 1 do
		local LAST_BAILIFF = CASTLE_BAILIFF_THRESHOLDS[3]
		local function est_eligible_rounds(B)
			if B < LAST_BAILIFF then
				return logisticfn(B-i) + est_eligible_rounds(B+1.25)
			else
				return 0
			end
		end
		local bldg = state.buildings[i]
		if bldg.owner then
			local r = player_r[bldg.owner]
			if bldg.ref.id == "residence" then
				r.residences = (r.residences or 0) + 1
			else
				r["bldg_"..bldg.ref.id] = est_eligible_rounds(state.bailiff)
			end
			if bldg.ref.id == "architect" then
				num_architects = num_architects + 1
				if num_architects > 1 then
					--this player's architect is NOT the
					--first architect on the road
					r["bldg_architect"] = r["bldg_architect"] / 4
				end
			end
		end
	end

	--castle contributions THIS phase
	do
		local start = 1
		if state.castle_phase > 1 then
			start = CASTLE_CONTRIBUTION_THRESHOLDS[state.castle_phase-1] + 1
		end
		for i = start, state.contribution_count do
			local pid = state.contributions[i]
			if pid and player_r[pid] then
				local r = player_r[pid]
				r.contrib_this_round = 1
			end
		end
	end

	-- PART ONE OF THREE
	-- for each building that has workers in it, quickly process the
	-- effect of that building, so that a proper valuation of those
	-- workers can be made.
	-- For buildings after the bridge, save a snapshot of the players
	-- resources at that point, so we can make a proper estimation of
	-- the value of workers given a provost position.

	for i = 1, CASTLE_POSITION - 1 do
		local bldg = state.buildings[i]
		local bldgref = bldg.ref
		process_building(bldg, player_r, rounds_left)
		if i == BRIDGE_POSITION + 1
			or i == state.provost
			or (i > BRIDGE_POSITION and bldg.occupants[1])
		then
			local tmp = {}
			for pid,pr in pairs(player_r) do
				tmp[pid] = resource_bag.clone(pr)
			end
			process_castle_and_round_end(tmp, rounds_left)
			process_future_carpenters(tmp, rounds_left)
			process_future_masons(tmp, rounds_left)
			process_future_architects(tmp, rounds_left)
			player_r_by_position[i] = tmp
		end
	end

	--PART TWO OF THREE
	-- Now, for each snapshot we made (stored in player_r_by_position),
	-- calculate the value of that player's goods.

	local function decaying_rewards2(rsrc_type)
		local p = ai.TUNABLE.VALUE_OF_RESOURCE[rsrc_type]
		local r1 = portion_of_game_left   --largest in first round
		local r2 = 1-r1                   --largest in last round
		return function(x)
				return
				--first 3 terms dominate at beginning of game
				  p[1] * r1 * logisticfn(1.6*x)
				+ p[2] * r1 * logisticfn(0.4*x)
				+ p[3] * r1 * logisticfn(0.1*x)
				- (p[1]+p[2]+p[3])/2 * r1
				--next 3 terms dominate at end of game
				+ p[4] * r2 * logisticfn(1.6*x)
				+ p[5] * r2 * logisticfn(0.4*x)
				+ p[6] * r2 * logisticfn(0.1*x)
				- (p[4]+p[5]+p[6])/2 * r2
			end
	end

	local value_of_money = decaying_rewards2("money")
	local value_of_food = decaying_rewards2("food")
	local value_of_wood = decaying_rewards2("wood")
	local value_of_stone = decaying_rewards2("stone")
	local value_of_cloth = decaying_rewards2("cloth")
	local value_of_gold = decaying_rewards2("gold")

local function check_alternates(r, v)
	local a = r._alternate
	while a do
		--pick the best choice from this alternates
		local best
		local best_bag
		local best_value = -math.huge
		for i,bag in ipairs(a.data) do
			local vv = {
				food = value_of_food(r.food + bag.food),
				wood = value_of_wood(r.wood + bag.wood),
				stone = value_of_stone(r.stone + bag.stone),
				cloth = value_of_cloth(r.cloth + bag.cloth),
				money = value_of_money(r.money + bag.money),
				}
			local value = vv.food + vv.wood + vv.stone + vv.cloth + vv.money
			if best_value < value then
				best_value = value
				best_bag = bag
				best = vv
			end
		end

		assert(best)
		v = best
		r.food = r.food + best_bag.food
		r.wood = r.wood + best_bag.wood
		r.stone = r.stone + best_bag.stone
		r.cloth = r.cloth + best_bag.cloth
		r.money = r.money + best_bag.money
		a = a.next
	end
	return v
end
	for position,pr in pairs(player_r_by_position) do
		for pid,r in pairs(pr) do
			local v = {
				food = value_of_food(r.food),
				wood = value_of_wood(r.wood),
				stone = value_of_stone(r.stone),
				cloth = value_of_cloth(r.cloth),
				money = value_of_money(r.money),
				}
			v = check_alternates(r, v)
			local combos = {
				gold_stone_stone = (r.gold >= 1 and r.stone >= 2 and 1 or 0),
				}
			r.est_score = r.score
				+ r.turn_order_position * ai.TUNABLE.VALUE_OF_ORDER
				+ v.food
				+ v.wood
				+ v.stone
				+ v.cloth
				+ value_of_gold(r.gold)
				+ v.money
				+ (r.contrib_this_round or 0) * ai.TUNABLE.VALUE_OF_FIRST_BLOCK
				+ r.contrib_now * ai.TUNABLE.VALUE_OF_CONTRIBUTION
				+ r.contrib_later * ai.TUNABLE.VALUE_OF_FUTURE_CONTRIBUTION
				+ r.favors * ai.TUNABLE.VALUE_OF_FAVOR
				+ r.income * rounds_left * ai.TUNABLE.VALUE_OF_INCOME_PER_ROUND
				+ rounds_left * ai.TUNABLE.VALUE_PER_ROUND
				+ 1 * ai.TUNABLE.VALUE_OFFSET
				+ (r.can_build_prestige or 0) * ai.TUNABLE.VALUE_OF_ARCHITECT
				+ r.vp_favor * ai.TUNABLE.VALUE_OF_VP_FAVORS
				+ r.money_favor * ai.TUNABLE.VALUE_OF_MONEY_FAVORS
				+ r.resource_favor * ai.TUNABLE.VALUE_OF_RESOURCE_FAVORS
				+ r.build_favor * ai.TUNABLE.VALUE_OF_BUILD_FAVORS
				+ r.vp_favor * rounds_left * ai.TUNABLE.VALUE_OF_VP_FAVORS_PER_ROUND
				+ r.money_favor * rounds_left * ai.TUNABLE.VALUE_OF_MONEY_FAVORS_PER_ROUND
				+ r.resource_favor * rounds_left * ai.TUNABLE.VALUE_OF_RESOURCE_FAVORS_PER_ROUND
				+ r.build_favor * rounds_left * ai.TUNABLE.VALUE_OF_BUILD_FAVORS_PER_ROUND
				+ (r.in_inn or 0) * ai.TUNABLE.VALUE_OF_INN
				+ ((r.residences or 0) >= 1 and 1 or 0) * ai.TUNABLE.VALUE_FIRST_RESIDENCE
				+ (r.residences or 0) * ai.TUNABLE.VALUE_PER_RESIDENCE
				+ combos.gold_stone_stone * ai.TUNABLE.VALUE_OF_COMBO.gold_stone_stone

			for i,bldg_id in ipairs(ALL_BUILDING_IDS) do
				--expected number of rounds owning the building
				local k1 = "bldg_"..bldg_id
				--probability of building this building in the future
				local k2 = "can_build_"..bldg_id
				r.est_score = r.est_score
					+ (r[k1] or 0) * ai.TUNABLE.VALUE_OF_BUILDING[bldg_id]
					+ (r[k2] or 0) * ai.TUNABLE.VALUE_OF_BUILDING_POTENTIAL[bldg_id]
			end
		end
	end


	-- PART THREE OF THREE
	-- Calculate the probability of the provost being at each particular
	-- position, and use that to calculate the expected value of each
	-- player's resources.

	local est_scores = {}
	for pid in pairs(state.players) do
		est_scores[pid] = 0
	end

	if state.phase == PHASE_WORKER_PLACEMENT
		or (state.phase == PHASE_ACTIVATE_BUILDINGS
			and state.active_building <= BRIDGE_POSITION
		   )
	then

	--record, for each position we made a snapshot, the probability
	--of that position (or later) being the true position
	for position = BRIDGE_POSITION + 1, CASTLE_POSITION - 1 do
		if player_r_by_position[position] then
			local pr = 1
			if position > BRIDGE_POSITION + 1 then
				pr = logisticfn(state.provost - position)
			end
			player_r_by_position[position]._pr = pr
		end
	end
	else
		--we know where the provost ended up, for sure
		local pr = player_r_by_position[state.provost]
		pr._pr = 1
		player_r_by_position = { [state.provost] = pr }

	end --end if provost future is uncertain

	--now, working backwards, use the marginal probability of each
	--position to calculate the player's _expected_ score.
	local last_pr = 0
	for position = CASTLE_POSITION - 1, BRIDGE_POSITION + 1, -1 do
		if player_r_by_position[position] then
			local player_r = player_r_by_position[position]
			player_r._probability = player_r._pr - last_pr
			last_pr = player_r._pr

			for pid in pairs(est_scores) do
				est_scores[pid] = est_scores[pid] + player_r[pid].est_score * player_r._probability
			end
--print(string.format("provost[%d] (Pr=%.3f) player 2 est score is %.1f",
--		position, player_r._probability, player_r[2].est_score))
--print(string.format("  food=%d, wood=%d", player_r[2].food, player_r[2].wood))
		end
	end
--print(string.format("player 2 aggregated est score is %.1f", est_scores[2]))

	if scratch_area then
		scratch_area.by_provost_position = player_r_by_position
		scratch_area.portion_of_game_left = portion_of_game_left
		scratch_area.rounds_left = rounds_left
	end

	if ai.VERBOSE then
		local pr = player_r_by_position[state.provost]
		local r = pr[current_turn()]
		for k,v in pairs(r) do
			if k ~= "_potential" then
			print(string.format("%-30s%6.1f", k, v))
			end
		end
	end

	return est_scores
end

function ai.learn_final_scores(real_scores)
	local scratch = {}
	local predicted = ai.estimate_final_scores(scratch)

	local adjustments = {
		}
	local function adjust(tunable_name, x, e)
		adjustments[tunable_name] = (adjustments[tunable_name] or 0)
			+ x * e
	end
	local function adjust_res(res_name, x, e)
		local tune_struct = ai.TUNABLE.VALUE_OF_RESOURCE[res_name]

		local r1 = scratch.portion_of_game_left
		local r2 = 1 - scratch.portion_of_game_left

		adjust("VALUE_OF_RESOURCE."..res_name..".1",
			r1 * logisticfn(1.6*x), e)
		adjust("VALUE_OF_RESOURCE."..res_name..".2",
			r1 * logisticfn(0.4*x), e)
		adjust("VALUE_OF_RESOURCE."..res_name..".3",
			r1 * logisticfn(0.1*x), e)
		adjust("VALUE_OF_RESOURCE."..res_name..".4",
			r2 * logisticfn(1.6*x), e)
		adjust("VALUE_OF_RESOURCE."..res_name..".5",
			r2 * logisticfn(0.4*x), e)
		adjust("VALUE_OF_RESOURCE."..res_name..".6",
			r2 * logisticfn(0.1*x), e)
	end

	for position,player_r in pairs(scratch.by_provost_position) do

		for pid in pairs(real_scores) do
			local error_amt = real_scores[pid] - predicted[pid]
			error_amt = error_amt * player_r._probability

			local r = player_r[pid]
			local combos = {
				gold_stone_stone = (r.gold >= 1 and r.stone >= 2 and 1 or 0),
				}
			adjust_res("food", r.food, error_amt)
			adjust_res("wood", r.wood, error_amt)
			adjust_res("stone", r.stone, error_amt)
			adjust_res("cloth", r.cloth, error_amt)
			adjust_res("gold", r.gold, error_amt)
			adjust_res("money", r.money, error_amt)
			adjust("VALUE_OF_CONTRIBUTION", r.contrib_now, error_amt)
			adjust("VALUE_OF_FUTURE_CONTRIBUTION", r.contrib_later, error_amt)
			adjust("VALUE_OF_ORDER", r.turn_order_position, error_amt)
			adjust("VALUE_OF_FAVOR", r.favors, error_amt)
			adjust("VALUE_OF_INCOME_PER_ROUND", r.income * scratch.rounds_left, error_amt)
			adjust("VALUE_PER_ROUND", scratch.rounds_left, error_amt)
			adjust("VALUE_OFFSET", 1, error_amt)
			adjust("VALUE_OF_ARCHITECT", (r.can_build_prestige or 0), error_amt)
			adjust("VALUE_OF_VP_FAVORS", r.vp_favor, error_amt)
			adjust("VALUE_OF_MONEY_FAVORS", r.money_favor, error_amt)
			adjust("VALUE_OF_RESOURCE_FAVORS", r.resource_favor, error_amt)
			adjust("VALUE_OF_BUILD_FAVORS", r.build_favor, error_amt)
			adjust("VALUE_OF_VP_FAVORS_PER_ROUND", r.vp_favor * scratch.rounds_left, error_amt)
			adjust("VALUE_OF_MONEY_FAVORS_PER_ROUND", r.money_favor * scratch.rounds_left, error_amt)
			adjust("VALUE_OF_RESOURCE_FAVORS_PER_ROUND", r.resource_favor * scratch.rounds_left, error_amt)
			adjust("VALUE_OF_BUILD_FAVORS_PER_ROUND", r.build_favor * scratch.rounds_left, error_amt)
			adjust("VALUE_OF_FIRST_BLOCK", (r.contrib_this_round or 0), error_amt)
			adjust("VALUE_OF_INN", (r.in_inn or 0), error_amt)
			adjust("VALUE_FIRST_RESIDENCE", ((r.residences or 0) >= 1 and 1 or 0), error_amt)
			adjust("VALUE_PER_RESIDENCE", (r.residences or 0), error_amt)
			adjust("VALUE_OF_COMBO.gold_stone_stone", combos.gold_stone_stone, error_amt)

			for i,bldg_id in ipairs(ALL_BUILDING_IDS) do
				local k = "bldg_"..bldg_id
				adjust("VALUE_OF_BUILDING."..bldg_id,
					(r[k] or 0), error_amt)
				local k = "can_build_"..bldg_id
				adjust("VALUE_OF_BUILDING_POTENTIAL."..bldg_id,
					(r[k] or 0), error_amt)
			end
		end
	end


	return adjustments
end

function ai.gather_stats(stats)

	stats.rounds = (stats.rounds or 0) + 1
	for i = 1,CASTLE_POSITION do
		if state.buildings[i].owner then
			local k1 = "bldg_"..state.buildings[i].ref.id
			stats[k1] = (stats[k1] or 0) + 1

			local k2 = "bldg_"..state.buildings[i].ref.id.."_taken"
			stats[k2] = (stats[k2] or 0) +
				(state.buildings[i].occupants[1] and 1 or 0)
		end
	end

end

return ai
